'''
Provides windows for viewing and editing data arrays in tao
'''
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox
from tkinter import filedialog

from .tao_widget import *
from pytao.util.parameters import str_to_tao_param, tao_parameter_dict
from .tao_set import *
from .tao_base_windows import *
from .tao_lat_windows import tao_ele_browser
#-----------------------------------------------------
# d2_data window

class tao_d2_data_window(tao_list_window):

    def __init__(self, root, pipe, *args, **kwargs):
        self.tao_id = 'data'
        tao_list_window.__init__(self, root, "Data", *args, **kwargs)
        self.pipe = pipe
        univ_list = self.pipe.cmd_in("python super_universe")
        n_universe = str_to_tao_param(univ_list.splitlines()[0])
        self.univ_frame = tk.Frame(self)
        tk.Label(self.univ_frame, text="Universe: ").grid(row=0,column=0,sticky="E")
        self.u_ix = tk.StringVar()
        u_ix_list = []
        for i in range(n_universe.value):
            u_ix_list.append(str(i+1))
        self.u_ix.set(u_ix_list[0])
        u_ix_box = tk.OptionMenu(
                self.univ_frame, self.u_ix, *u_ix_list, command=self.refresh)
        u_ix_box.grid(row=0,column=1,sticky="W")
        self.univ_frame.pack(fill="both", expand=0)

        # Populate self.list_frame
        self.refresh()


    def refresh(self):
        '''
        Clears self.list_frame and fills it with the current universe's
        d2/d1 data
        '''
        # Clear self.list_frame:
        for child in self.list_frame.winfo_children():
            child.destroy()
        # Get this universe's d2_data
        u_ix = self.u_ix.get()
        d2_data_list = self.pipe.cmd_in("python data_d2_array " + u_ix)
        d2_data_list = d2_data_list.splitlines()
        if len(d2_data_list) == 0:
            tk.Label(self.list_frame, text="NO DATA FOR THIS UNIVERSE").pack()
            return
        for d2_data_item in d2_data_list:
            new_frame = d2_data_frame(
                    self.list_frame, self.root, self.pipe, d2_data_item, u_ix)
            new_frame.frame.pack()


#-----------------------------------------------------

class d2_data_frame():
    '''
    Sets up an object with the following attributes:
    self.frame = a tk frame that will hold all relevant widgets and labels
    self.name = holds the name of the d2_data this frame represents
    self.d1_data_list = list of all d1_data items contained by this d2_datum
    NOTE: master should be the frame that this frame is gridded into
    (usually a list_frame), while root should be the application root window
    '''
    def __init__(self, master, root, pipe, d2_data_name, u_ix):
        self.frame = tk.Frame(master)
        self.root = root
        self.name = d2_data_name
        self.d1_data_list = []
        self.d1_using_list = [] #holds using information
        self.d1_ix_lb_list = [] #holds index lower bounds
        self.d1_ix_ub_list = [] #holds index upper bounds
        d2_info = pipe.cmd_in("python data_d1_array " + u_ix + "@" + self.name)
        d2_info = d2_info.splitlines()
        for item in d2_info:
            item = item.split(';')
            self.d1_data_list.append(item[3])
            self.d1_using_list.append(item[4])
            self.d1_ix_lb_list.append(int(item[5]))
            self.d1_ix_ub_list.append(int(item[6]))
        tk.Label(self.frame, text=self.name).grid(row=0, column=0, columnspan=4, sticky='W')
        tk.Label(self.frame, text="Indices").grid(row=0, column=4)
        tk.Label(self.frame, text="Using").grid(row=0, column=5)
        for i in [0,4,5]:
            self.frame.grid_columnconfigure(i, pad=10)
        for i in range(len(self.d1_data_list)):
            tk.Label(self.frame, text=self.d1_data_list[i]).grid(row=i+1,column=0, sticky='W')
            tk.Button(self.frame, text="View...",
                    command=self.open_d1_callback(self.name, self.d1_data_list[i], pipe,
                        self.d1_ix_lb_list[i], self.d1_ix_ub_list[i], u_ix)).grid(
                                row=i+1,column=1)
            tk.Button(self.frame, text="Edit...",
                    command=self.edit_d2_callback(self.name, pipe)).grid(
                            row=i+1, column=2)
            mytext = str(self.d1_ix_lb_list[i]) + ":" + str(self.d1_ix_ub_list[i])
            tk.Label(self.frame, text=mytext).grid(row=i+1, column=4)
            tk.Label(self.frame, text=self.d1_using_list[i]).grid(row=i+1, column=5)

    def open_d1_callback(self, d2_data_name, d1_data_name,
            pipe, ix_lb, ix_ub, u_ix):
        return lambda : self.open_d1(
                d2_data_name, d1_data_name, pipe, ix_lb, ix_ub, u_ix)

    def edit_d2_callback(self, d2_data_name, pipe):
        return lambda : self.edit_d2(d2_data_name, pipe)

    def edit_d2(self, d2_data_name, pipe):
        ''' Opens the new data window and clones this d2_array in for editing'''
        win = tao_new_data_window(self.root, pipe, default=d2_data_name)

    def open_d1(self, d2_data_name, d1_data_name, pipe, ix_lb, ix_ub, u_ix):
        '''
        Opens a window with detailed information for d2_data_name.d1_data_name
        '''
        win = tao_d1_data_window(
                self.root, pipe, d2_data_name + '.' + d1_data_name, u_ix, ix_lb, ix_ub)

#-----------------------------------------------------------------
# d1_data window

class tao_d1_data_window(lw_table_window):
    '''
    With lw set to True, opens a lw_table_window instead
    '''
    def __init__(self, root, pipe, d1_data_name,
            u_ix, ix_lb, ix_ub, *args, **kwargs):
        self.tao_id = 'data'
        self.u_ix = u_ix
        self.ix_lb = ix_lb
        self.ix_ub = ix_ub
        self.lw = root.lw
        title_list = ["Index",
                "d1_data_name",
                "Merit type",
                "Ref Element",
                "Start Element",
                "Element Name",
                "Meas value",
                "Model value",
                "Design value",
                "Useit_opt",
                "Useit_plot",
                "good_user",
                "Weight",
                "Exists"]
        bulk_template = []
        bulk_template.append([str_to_tao_param("meas_value;REAL;T;"), 6])
        bulk_template.append([str_to_tao_param("good_user;LOGIC;T;"), 11])
        bulk_template.append([str_to_tao_param("weight;REAL;T;"), 12])

        bulk_set_format = "set data " + str(self.u_ix) + '@{}|'
        set_format = "set data " + str(self.u_ix) + '@{}[{}]|'
        if self.lw:
            lw_table_window.__init__(self, root, pipe, d1_data_name, title_list,
                    bulk_template, bulk_set_format, set_format, *args, **kwargs)
        else:
            table_window.__init__(self, root, pipe, d1_data_name, title_list,
                    bulk_template, bulk_set_format, set_format, *args, **kwargs)

    def refresh(self):
        self.d_list = self.pipe.cmd_in(
                "python data_d_array " + self.u_ix + '@' + self.array_name)
        self.d_list = self.d_list.splitlines()
        self.row_num = len(self.d_list)
        if self.lw:
            lw_table_window.refresh(self)
        else:
            table_window.refresh(self)

    def list_row_fetch(self, index):
        '''
        Returns row number index to be appended to self.list_rows
        '''
        return d1_data_list_entry(self.list_frame, self.d_list[index])

    def lw_list_row_fetch(self, index):
        '''
        For use with lw_table_window.  Returns a list of the values
        to be displayed in each column of row #index.
        '''
        return self.d_list[index].split(';')

    def lw_detail_callback(self, event=None):
        '''
        Callback for the light-weight table to open a detail window
        '''
        lw_table_window.open_detail_window_callback(self, event)

    def bulk_set(self, fill_choices, parent):
        '''
        Overload of lw_table_window.bulk_set that calls refresh data and plot windows
        '''
        lw_table_window.bulk_set(self, fill_choices, parent)
        # Refresh data-related windows
        for win in self.root.refresh_windows['data']:
            win.refresh()
        for win in self.root.refresh_windows['plot']:
            win.refresh()

    def open_detail_window(self, index):
        self.param_list = self.pipe.cmd_in("python data " + str(self.u_ix) + '@'
                + self.array_name + '[' + str(index) + ']')
        self.param_list = self.param_list.splitlines()
        for i in range(len(self.param_list)):
            self.param_list[i]=str_to_tao_param(self.param_list[i])
        if self.lw:
            lw_table_window.open_detail_window(self, index)
        else:
            table_window.open_detail_window(self, index)

    def detail_set_callback(self, tao_list, set_str):
        '''
        Overload of lw_table_window.detail_set_callback that refreshes
        data-related windows
        '''
        lw_table_window.detail_set_callback(self, tao_list, set_str)
        # Refresh data-related windows
        for win in self.root.refresh_windows['data']:
            win.refresh()
        for win in self.root.refresh_windows['plot']:
            win.refresh()

class tao_new_data_window(Tao_Toplevel):
    '''
    Provides a window for creating new d2 data arrays (and their associated d1 arrays)
    Pass the name of an existing d2_array to open that array and start editing its d1_arrays
    '''
    def __init__(self, root, pipe, default=None, *args, **kwargs):
        self.root = root
        Tao_Toplevel.__init__(self, root, *args, **kwargs)
        self.pipe = pipe
        self.rowconfigure(0, weight=1)
        self.title('New D2 Data')
        self.name = ""

        #Frame for inputting d2 parameters
        self.d2_frame = tk.Frame(self)
        self.d2_frame.grid(row=0, column=0, sticky='NSEW')

        #Frames for inputting d1 parameters
        self.d1_frame = tk.Frame(self)
        self.nb_exists = False #used to track if the ttk Notebook has been set up
        self.d1_frame_list = []

        self.fill_d2_frame()
        if default != None:
            self.d2_param_list[0].tk_var.set(default)
            self.load_d1_frame(ask=False)
            for d1_frame in self.d1_frame_list:
                d1_frame.fill_defaults()

    def fill_d2_frame(self):
        tk.Label(self.d2_frame, text="New D2 Data Structure",
                font=('Sans', 16, 'bold')).grid(
                        row=0, column=0, columnspan=2, sticky='EW')

        # Small bit of setup
        self.uni = tk.StringVar()
        self.uni.set("All")
        uni_max = self.pipe.cmd_in("python super universe")
        uni_max = str_to_tao_param(uni_max.splitlines()[0]).value
        uni_list = ["All"] + list(range(1, uni_max+1))
        def my_ttp(x):
            ''' shorcut for the following commonly used construct '''
            return tk_tao_parameter(str_to_tao_param(x), self.d2_frame, self.pipe)

        # Clone existing d2
        existing_d2_arrays = self.pipe.cmd_in("python data_d2_array 1")
        existing_d2_arrays = ['None'] + existing_d2_arrays.splitlines()
        self.clone_d2 = tk.StringVar()
        self.clone_d2.set('None')

        # Widgets
        self.d2_param_list = [my_ttp("name;STR;T;"),
                tk.OptionMenu(self.d2_frame, self.uni, *uni_list),
                my_ttp("data_source;ENUM;T;"),
                my_ttp("data^merit_type;ENUM;T;"),
                my_ttp("weight;REAL;T;"),
                my_ttp("good_user;LOGIC;T;T"),
                tk.OptionMenu(self.d2_frame, self.clone_d2, *existing_d2_arrays)]
        # Index loop-up table
        self.d2_array_indices = {}
        for d2_param, index in zip(self.d2_param_list, range(len(self.d2_param_list))):
            if isinstance(d2_param, tk_tao_parameter):
                self.d2_array_indices[d2_param.param.name.split('^')[-1]] = index

        # Labels
        self.d2_label_list = [tk.Label(self.d2_frame, text="d2_data Name:"),
                tk.Label(self.d2_frame, text="Which universe:"),
                tk.Label(self.d2_frame, text="Default data source:"),
                tk.Label(self.d2_frame, text="Default merit type:"),
                tk.Label(self.d2_frame, text="Default weight:"),
                tk.Label(self.d2_frame, text="Default good_user:"),
                tk.Label(self.d2_frame, text="Clone existing d2_array:")]

        # Grid widgets and labels
        for i in range(len(self.d2_param_list)):
            self.d2_label_list[i].grid(row=i+1, column=0, sticky='W')
            if i in [1, 6]: #non tk_tao_parameters
                self.d2_param_list[i].grid(row=i+1, column=1, sticky='EW')
            else:
                self.d2_param_list[i].tk_wid.grid(row=i+1, column=1, sticky='EW')

        # Warning labels
        self.name_warning_empty = tk.Label(self.d2_frame, text="Cannot be empty")
        # Next button
        self.next_b = tk.Button(self.d2_frame, text="Next", command=self.load_d1_frame)
        self.next_b.grid(row=i+2, column=1, sticky='W')

        # Focus the name entry
        self.d2_param_list[0].tk_wid.focus_set()

    def load_d2_frame(self):
        self.d1_frame.pack_forget()
        self.d2_frame.grid(row=0, column=0, sticky='NSEW')

    def load_d1_frame(self, ask=True):
        '''
        Ungrids self.d2_frame, grids self.d1_frame, and sets up a notebook
        for the d1_frames if necessary.
        Set ask=False to skip message boxes
        '''
        clone_dict = {} # keys=d2_array names, values=lists of d1s
        # Check if d2_name is nonempty
        name = self.d2_param_list[0].tk_var.get().strip()
        if name == "":
            self.name_warning_empty.grid(row=1, column=2, sticky='W')
            return

        # Conditions
        c1 = (name in self.pipe.cmd_in('python data_d2_array 1').splitlines()) #name in use
        c2 = (self.clone_d2.get() != 'None') #clone has been specified
        c3 = (self.name != name) #name has changed
        c4 = False #discard existing work
        ans_var = tk.StringVar()
        # Ask if user wants to keep existing data
        if c3 and self.name:
            self.name = name
            if ask:
                tao_message_box(self.root, self, ans_var, title='Warning', message='Would you like to keep or discard the d1_arrays you defined for ' + self.name + '?', choices=['Keep', 'Discard'])
            else:
                ans_var.set('Discard')
            if ans_var.get() == 'Keep':
                c4 = False
            elif ans_var.get() == 'Discard':
                c4 = True
            else:
                return
        # Ask if user wants to load existing data
        if c1:
            if ask:
                ans = messagebox.askyesno('Warning', name + " already exists as a d2 data array.    Would you like to clone its existing d1_arrays?", parent=self)
            else:
                ans = True
            if ans:
                # Will need to read in data for existing d2
                clone_dict[name] = self.pipe.cmd_in('python data_d1_array ' + name).splitlines()
        # Clone existing data
        if c2 and (self.clone_d2.get() != name):
            clone_dict[self.clone_d2.get()] = self.pipe.cmd_in('python data_d1_array ' + self.clone_d2.get()).splitlines()
        # Remake d1_frames and notebook if necessary
        if c4 and self.nb_exists:
            for frame in self.d1_frame_list:
                frame.destroy()
            self.d1_frame_list = []
            self.notebook.destroy()
            self.new_tab_frame.destroy()
            self.back_b.destroy()
            self.create_b.destroy()
            self.nb_exists = False
        else:
            self.name = name
            self.name_warning_empty.grid_forget()

        # Possibly create self.notebook
        if not self.nb_exists:
            self.notebook = ttk.Notebook(self.d1_frame)
            self.notebook.pack(side='top', fill='both', expand=1)
            self.nb_exists = True

            self.d1_index = 0 #marks current tab index

            # New tab button
            self.new_tab_frame = tk.Frame(self.notebook)
            self.notebook.insert('end', self.new_tab_frame)
            self.notebook.tab(len(self.d1_frame_list), text='+')
            self.notebook.bind('<<NotebookTabChanged>>', self.tab_handler)

            # Back button
            self.back_b = tk.Button(self.d1_frame, text="Back", command=self.load_d2_frame)
            self.back_b.pack(side='left')

            # Create button
            self.create_b = tk.Button(self.d1_frame, text="Create!", command=self.create_data)
            self.create_b.pack(side='right')

        self.d2_frame.grid_forget()
        self.d1_frame.pack(fill='both', expand=1)
        self.title("New D1 Data Structures of: " + self.d2_param_list[0].tk_var.get())

        # Clone the requested d2 array(s)
        if clone_dict != {}:
            # Progress window
            num_bars = 0
            for k in clone_dict.keys():
                num_bars += len(clone_dict[k])
            self.pw = tao_progress_window(self.root, self, num_bars)
            self.pw.title("Loading...")
            tk.Label(self.pw, text='Loading existing d1_arrays').grid(row=0, column=0, columnspan=2, sticky='EW')
            # Find correct universe number
            if self.uni.get() == "All":
                u = "1"
            else:
                u = self.uni.get()
            if name in clone_dict.keys():
                for d1 in clone_dict[name]:
                    self.add_d1_frame(d1, u+'@'+name)
            if (self.clone_d2.get() in clone_dict.keys()) and (name != self.clone_d2.get()):
                for d1 in clone_dict[self.clone_d2.get()]:
                    self.add_d1_frame(d1, u+'@'+self.clone_d2.get())
            self.notebook.select(0)
            self.pw.destroy()

        # Copy d2 defaults into d1_arrays
        for i in range(len(self.d1_frame_list)):
            val = self.d2_param_list[2].tk_var.get()
            self.d1_frame_list[i].d1_array_wids[1].tk_var.set(val)
            val = self.d2_param_list[3].tk_var.get()
            self.d1_frame_list[i].d1_array_wids[3].tk_var.set(val)
            val = self.d2_param_list[4].tk_var.get()
            self.d1_frame_list[i].d1_array_wids[4].tk_var.set(val)

    def create_data(self, event=None):
        '''
        Takes the information from the d2_frame and d1_frames and runs
        the necessary commands to create the data in tao, then closes
        the create data window
        '''
        # Input validation
        messages = []
        for d1_frame in self.d1_frame_list:
            # Check names
            if not d1_frame.name_handler():
                name_m = "Please check d1_array names."
                if name_m not in messages:
                    messages.append(name_m)
            if not d1_frame.ix_min_handler(strict=True):
                messages.append("Please check the base index for " + d1_frame.name)
            if not d1_frame.length_handler(strict=True):
                messages.append("Please check the array length for " + d1_frame.name)
            if d1_frame.d1_array_wids[2].tk_var.get() == "":
                messages.append("Please choose a data type for " + d1_frame.name)
            # Check for semicolons in any fields
            semi_message = "Semicolons not allowed in any input field"
            caret_message = "Carets not allowed in any input field"
            broken = False #Used to break out of the below for loops
            # Check for semicolons/carets
            for ttp in d1_frame.d1_array_wids:
                if str(ttp.tk_var.get()).find(';') != -1:
                    messages.append(semi_message)
                    broken = True
                    break
                if str(ttp.tk_var.get()).find('^') != -1:
                    messages.append(caret_message)
                    broken = True
                    break
            for data_dict in d1_frame.data_dict.values():
                if broken:
                    break
                for d in data_dict.values():
                    if str(d).find(';') != -1:
                        messages.append(semi_message)
                        broken = True
                        break
                    if str(d).find('^') != -1:
                        messages.append(caret_message)
                        broken = True
                        break
                if broken:
                    break
        for m in messages:
            messagebox.showwarning("Error", m, parent=self)
        if messages != []:
            return
        # Book-keeping
        datum_params = ['data_type', 'ele_ref_name', 'ele_start_name', 'ele_name',
                'data^merit_type', 'meas_value', 'good_meas', 'ref_value', 'good_ref',
                'weight', 'good_user', 'data_source', 'eval_point', 's_offset',
                '1^ix_bunch', 'invalid_value', 'spin_n0_x', 'spin_n0_y', 'spin_n0_z']
        d1_params = ['name', 'data_source', 'data_type', 'data^merit_type', 'weight', 'eval_point', 'good_user']
        d2_params = ['name', 'uni', 'data_source', 'data^merit_type', 'weight', 'good_user']
        # Create the data array
        if self.uni.get() == 'All':
            uni_max = self.pipe.cmd_in("python super universe")
            uni_max = str_to_tao_param(uni_max.splitlines()[0]).value
            uni_list = list(range(1, uni_max+1))
        else:
            uni_list = [int(self.uni.get())]
        d1_count = 0 # used to make the corrent number of progress bars
        for u in uni_list:
            cmd_str = 'python data_d2_create ' + str(u) + '@' + self.name
            cmd_str += '^^' + str(len(self.d1_frame_list)) + '^^'
            for d1_frame in self.d1_frame_list:
                d1_count += 1
                # min/max indices for each d1_array
                cmd_str += str(d1_frame.name) + '^^'
                cmd_str += str(d1_frame.ix_min) + '^^'
                cmd_str += str(d1_frame.ix_max) + '^^'
            # Create the d2/d1_arrays
            self.pipe.cmd_in(cmd_str)
        # Progress bars
        self.pw = tao_progress_window(self.root, self, d1_count)
        for u in uni_list:
            # Progress window config
            for d1_frame in self.d1_frame_list:
                self.pw.label_vars[self.pw.ix].set(
                        'Creating' + str(u) + '@' + self.name + '.' + d1_frame.name)
                self.pw.set_max(self.pw.ix, d1_frame.ix_max-d1_frame.ix_min+1)
                # set individual data parameters
                for j in range(d1_frame.ix_min, d1_frame.ix_max+1):
                    self.pw.set_val(self.pw.ix, j-d1_frame.ix_min)
                    cmd_str = 'python datum_create '
                    cmd_str += str(u) + '@' + self.name + '.' + d1_frame.name + '[' + str(j) + ']'
                    for p in datum_params:
                        #look in d1_frame.data_dict
                        if (j in d1_frame.data_dict.keys()) and (p in d1_frame.data_dict[j].keys()):
                            value = d1_frame.data_dict[j][p]
                        elif p in d1_params:
                            value = d1_frame.d1_array_wids[d1_params.index(p)].tk_var.get()
                        elif p in d2_params:
                            value = self.d2_param_list[d2_params.index(p)].tk_var.get()
                        else:
                            value = ""
                        if p in ["good_user", "good_meas", "good_ref"]: # replace with T or F
                            value = "T" if value else "F"
                        cmd_str += '^^' + value
                    self.pipe.cmd_in(cmd_str)
                self.pw.ix += 1
        # Recalculate for the current universe
        self.pipe.cmd_in('set universe -1 recalculate')
        # Close the window
        self.destroy()
        # Refresh data-related windows
        for win in self.root.refresh_windows['data']:
            win.refresh()
        for win in self.root.refresh_windows['plot']:
            win.refresh()

    def tab_handler(self, event=None):
        '''
        Handles new tab creation and updates self.d1_index as necessary
        '''
        # Check if the new tab frame has been selected
        if self.notebook.select() == self.new_tab_frame._w:
            # Add new tab
            self.d1_frame_list.append(new_d1_frame(self))
            self.d1_index = len(self.d1_frame_list)-1
            self.notebook.insert(self.d1_index, self.d1_frame_list[-1])
            self.notebook.tab(self.d1_index, text=self.d1_frame_list[-1].BLANK_TITLE)
            self.notebook.select(self.d1_index)
        else:
            # Update self.d1_index
            for i in range(len(self.d1_frame_list)):
                frame = self.d1_frame_list[i]
                if self.notebook.select() == frame._w:
                    self.d1_index = i
                    # Unblock frame's handlers
                    frame.handler_block = False
                    break

    def add_d1_frame(self, d1_name, d2_name, event=None):
        '''
        Creates a new d1_frame for d1_name and reads existing data in from Tao if present
        Takes a line of output from python data_d1_array for d1_name
        d2_name should be "uni@d2_name"
        '''
        d1 = d1_name.split(';')[3]
        self.d1_frame_list.append(
                new_d1_frame(self, name=d1, full_name=d2_name + '.' + d1))
        ix = len(self.d1_frame_list) - 1
        self.notebook.insert(ix, self.d1_frame_list[-1])
        self.notebook.tab(ix, text=d1)

    def refresh(self):
        '''
        Only here in case something tries to refresh this window
        '''
        pass

class new_d1_frame(tk.Frame):
    '''
    Provides a frame for inputting properties of a d1_data array.
    To load an existing d1_array, pass the d1_array name as name
    (e.g. x,y, NOT orbit.x, orbit.y)
    Also pass the full name that should be used for data lookup
    (e.g. 1@orbit.x)
    '''
    def __init__(self, d2_array, name="", full_name=""):
        tk.Frame.__init__(self, d2_array.notebook)
        for i in list(range(3)) + list(range(4, 6)):
            self.grid_columnconfigure(i, weight=1)
        self.d2_array = d2_array
        self.pipe = self.d2_array.pipe
        self.handler_block = False
        self.BLANK_TITLE = "[UNTITLED]"
        if name == "":
            self.name = self.BLANK_TITLE #Default
        else:
            self.name = name

        # d1 Widgets
        def d1_ttp(x):
            ''' Shortcut for commonly used construct '''
            return tk_tao_parameter(str_to_tao_param(x), self, self.pipe)

        self.d1_array_wids = [d1_ttp('name;STR;T;'),
                d1_ttp('data_source;ENUM;T;'),
                d1_ttp('data_type;DAT_TYPE;T;'),
                d1_ttp('data^merit_type;ENUM;T;'),
                d1_ttp('weight;REAL;T;'),
                d1_ttp('eval_point;ENUM;T;T'),
                d1_ttp('good_user;LOGIC;T;T'),
                d1_ttp('ix_min;INT;T;'),
                d1_ttp('ix_max;INT;T;')]
        # Look-up table of array indices
        self.d1_array_indices = {}
        for d1_wid, index in zip(self.d1_array_wids, range(len(self.d1_array_wids))):
            self.d1_array_indices[d1_wid.param.name.split('^')[-1]] = index
        # d1 labels (NAMES AS STRINGS ONLY)
        self.d1_array_labels = ["d1_array Name:", "Default data source:",
                "Default data type:", "Default merit type:", "Default weight:",
                "Default eval point:", "Default good_user:", "Base index:", "Array length:"]
        # Read in defaults from d2 level
        for d1_name, d1_ix in self.d1_array_indices.items():
            if d1_name != "name":
                if d1_name in self.d2_array.d2_array_indices.keys():
                    d2_ix = self.d2_array.d2_array_indices[d1_name]
                    val = self.d2_array.d2_param_list[d2_ix].tk_var.get()
                    self.d1_array_wids[d1_ix].tk_var.set(val)
        #val = self.d2_array.d2_param_list[2].tk_var.get()
        #self.d1_array_wids[1].tk_var.set(val)
        #val = self.d2_array.d2_param_list[3].tk_var.get()
        #self.d1_array_wids[3].tk_var.set(val)
        #val = self.d2_array.d2_param_list[4].tk_var.get()
        #self.d1_array_wids[4].tk_var.set(val)
        #val = self.d2_array.d2_param_list[5].tk_var.get()
        #self.d1_array_wids[6].tk_var.set(val)

        # Set eval point to End by default
        self.d1_array_wids[self.d1_array_indices["eval_point"]].tk_var.set("End")
        # Grid widgets and labels:
        tk.Label(self, text=self.d1_array_labels[0]).grid(row=3, column=0, sticky='W')
        self.d1_array_wids[0].tk_wid.grid(row=3, column=1, sticky='EW')
        for i in range(1, len(self.d1_array_wids)):
            tk.Label(self, text=self.d1_array_labels[i]).grid(row=i+4, column=0, sticky='W')
            self.d1_array_wids[i].tk_wid.grid(row=i+4, column=1, sticky='EW')
        i = i+4
        # Set name
        if self.name != self.BLANK_TITLE:
            self.d1_array_wids[0].tk_var.set(self.name)

        # Warning labels
        # (defined here to be gridded/ungridded as necessary)
        self.name_warning_empty = tk.Label(self, text="Must not be empty")
        self.name_warning_invalid = tk.Label(self, text="d1 name already in use")
        self.ix_min_warning = tk.Label(self, text="Must be a non-negative integer")
        self.length_warning = tk.Label(self, text="Must be a positive integer")

        # Grid settings for the above warning labels (use warning_label.grid(**settings) to
        # unroll the appropriate dictionary into the keyword settings)
        self.name_warning_gs = {"row" : 2, "column" : 2, "sticky" : 'W'}
        self.ix_min_warning_gs = {"row" : 8, "column" : 2, "sticky" : 'W'}
        self.length_warning_gs = {"row" : 9, "column" : 2, "sticky" : 'W'}

        # Responses to edits
        self.d1_array_wids[self.d1_array_indices["name"]].tk_wid.bind('<FocusOut>', self.name_handler)
        self.d1_array_wids[self.d1_array_indices["ix_min"]].tk_wid.bind('<FocusOut>', self.ix_min_handler)
        self.d1_array_wids[self.d1_array_indices["ix_max"]].tk_wid.bind('<FocusOut>', self.length_handler)
        self.d1_array_wids[self.d1_array_indices["data_source"]].tk_var.trace('w', self.data_source_handler)
        self.d1_array_wids[self.d1_array_indices["data_type"]].tk_var.trace('w', self.data_type_handler)


        # "Fill" to datums
        def datum_fill_callback(ix):
            '''Helper function for buttons below'''
            return lambda : self.datum_fill(ix)
        for ix in range(1, 6):
            tk.Button(self, text="Fill to datums", command=datum_fill_callback(ix)).grid(row=ix+4, column=2, sticky='EW')

        ttk.Separator(self, orient='horizontal').grid(row=i+1, column=0, columnspan=3, sticky='EW')
        i = i+1

        # Element browsers
        tk.Label(self, text="Evaluation elements:").grid(row=4, column=0, sticky='W')
        self.ele_name_button = tk.Button(self, text="Browse...", command=lambda : self.lat_browser('name'))
        self.ele_name_button.grid(row=4, column=1, sticky='EW')

        tk.Label(self, text="Start elements:").grid(row=i+2, column=0, sticky='W')
        self.ele_start_name_button = tk.Button(self, text="Browse...", command=lambda : self.lat_browser('start_name'))
        self.ele_start_name_button.grid(row=i+2, column=1, sticky='EW')

        tk.Label(self, text="Ref elements:").grid(row=i+3, column=0, sticky='W')
        self.ele_ref_name_button = tk.Button(self, text="Browse...", command=lambda : self.lat_browser('ref_name'))
        self.ele_ref_name_button.grid(row=i+3, column=1, sticky='EW')

        #ttk.Separator(self, orient='horizontal').grid(row=i+4, column=0, columnspan=3, sticky='EW')
        #i = i+1

        # Individual data
        self.data_dict = {}
        self.datum_frame = tk.Frame(self)
        tk.Label(self, text="Datum:").grid(row=1, column=4, sticky='W')
        self.data_ix = tk.StringVar()
        self.ix_min = -1
        self.ix_max = -1
        self.length = -1
        self.data_chooser = ttk.Combobox(self, textvariable=self.data_ix,
                values=['PLEASE SPECIFY MIN/MAX INDICES'], state='readonly')
        self.data_chooser.bind('<<ComboboxSelected>>', self.make_datum_frame)
        self.data_chooser.grid(row=1, column=5, sticky='EW')
        # Set ix_min and ix_max for existing d1_arrays
        if self.name != self.BLANK_TITLE:
            ix_data = self.pipe.cmd_in('python data_d1_array ' + full_name)
            ix_data = ix_data.splitlines()
            for line in ix_data:
                if line.split(';')[3] == self.name:
                    break
            #line is now set to the relevant line
            ix_min = int(line.split(';')[5])
            ix_max = int(line.split(';')[6])
            length = ix_max - ix_min + 1
            self.d1_array_wids[-2].tk_var.set(str(ix_min))
            self.d1_array_wids[-1].tk_var.set(str(length))
            self.ix_min_handler()
            self.length_handler()
        # Fill self.data_dict for existing d1 arrays
        data_dict_params = ['data_source', 'data_type', 'ele_name', 'ele_start_name', 'ele_ref_name',
                'data^merit_type', 'meas_value', 'ref_value', 'weight', 'good_user', '1^ix_bunch',
                'eval_point', 's_offset']
        if self.name != self.BLANK_TITLE:
            self.d2_array.pw.label_vars[self.d2_array.pw.ix].set(
                    'Loading ' + self.d2_array.name + '.' + self.name + '...')
            self.d2_array.pw.set_max(self.d2_array.pw.ix, self.ix_max-self.ix_min+1)
            #self.d2_array.update_idletasks()
            for i in range(self.ix_min, self.ix_max+1):
                self.d2_array.pw.set_val(self.d2_array.pw.ix, i-self.ix_min)
                #self.d2_array.update_idletasks()
                self.data_dict[i] = {}
                existing_data = self.pipe.cmd_in(
                        'python data ' + full_name + '[' + str(i) + ']')
                existing_data = existing_data.splitlines()
                for p in data_dict_params:
                    # Find correct line
                    for line in existing_data:
                        if line.find(p + ';') == 0:
                            break
                    # Write to self.data_dict
                    self.data_dict[i][p] = line.split(';')[3].strip()
            self.d2_array.pw.ix += 1
            # Load first datum
            self.data_ix.set(self.ix_min)
            self.make_datum_frame()
        else:
            self.ix_min = 1
            self.d1_array_wids[self.d1_array_indices["ix_min"]].tk_var.set(1)
            self.data_ix.set(self.ix_min)
            self.make_datum_frame()

        # Delete button
        tk.Button(self, text="DELETE THIS D1_ARRAY", fg='red', command=self.delete).grid(
                row=1, column=0, columnspan=3, sticky='EW')

        # Duplicate button
        tk.Button(self, text="Duplicate this d1_array", command=self.duplicate).grid(
                row=2, column=0, columnspan=3, sticky='EW')

        # Vertical separator and titles
        ttk.Separator(self, orient="vertical").grid(row=0, column=3, rowspan=30, sticky='NS')
        tk.Label(self, text="D1_array Settings:", font="Sans 16").grid(row=0, column=0, sticky='W')
        tk.Label(self, text="Datum Settings:", font="Sans 16").grid(row=0, column=4, sticky='W')

        # Focus the d1 name widget
        self.d1_array_wids[0].tk_wid.focus_set()

    def fill_defaults(self, event=None):
        '''
        Copies the properties of the first datum and makes them the defaults at the d1 level
        '''
        self.d1_array_wids[1].tk_var.set(self.data_dict[self.ix_min]['data_source'])
        self.d1_array_wids[2].tk_var.set(self.data_dict[self.ix_min]['data_type'])
        self.d1_array_wids[3].tk_var.set(self.data_dict[self.ix_min]['data^merit_type'])
        self.d1_array_wids[4].tk_var.set(self.data_dict[self.ix_min]['weight'])
        self.d1_array_wids[5].tk_var.set(self.data_dict[self.ix_min]['eval_point'])
        self.d1_array_wids[6].tk_var.set(True if self.data_dict[self.ix_min]['good_user'] == 'T' else False)

    def delete(self, ask=True, event=None):
        '''
        Deletes this d1_array frame
        Call with ask = False to skip confirmation
        '''
        # Ask for confirmation
        if ask:
            ans = messagebox.askokcancel("Delete " + self.name, "Delete this d1_array and its associated data?", parent=self.d2_array)
            if not ans:
                return

        # Remove from parent d1_frame_list
        ix = self.d2_array.d1_index
        if ix > 0:
            # Prefer to move to the tab to the left
            self.d2_array.notebook.select(ix-1)
        self.d2_array.d1_frame_list.pop(ix)

        # Destroy self
        self.destroy()

    def duplicate(self, event=None):
        '''
        Adds a new d1_frame to self.d2_array.d1_frame_list that is a copy of
        this frame, and changes focus to that frame
        '''
        # Don't run any handlers for this d1_frame
        self.handler_block = True
        self.d2_array.d1_frame_list.append(new_d1_frame(self.d2_array))
        ix = len(self.d2_array.d1_frame_list) - 1
        #self.d2_array.d1_index = ix
        # Copy properties into new frame
        self.d2_array.d1_frame_list[-1].name = self.name + '_copy'
        self.d2_array.d1_frame_list[-1].ix_min = self.ix_min
        self.d2_array.d1_frame_list[-1].ix_max = self.ix_max
        self.d2_array.d1_frame_list[-1].data_dict = self.data_dict
        for i in range(len(self.d1_array_wids)):
            if i == 0:
                self.d2_array.d1_frame_list[-1].d1_array_wids[i].tk_var.set(self.d1_array_wids[i].tk_var.get() + '_copy')
            else:
                self.d2_array.d1_frame_list[-1].d1_array_wids[i].tk_var.set(self.d1_array_wids[i].tk_var.get())
        # Run all input validation handlers
        self.d2_array.notebook.insert(ix, self.d2_array.d1_frame_list[-1])
        self.d2_array.notebook.select(ix)
        self.d2_array.tab_handler()
        self.update_idletasks()
        self.d2_array.d1_frame_list[-1].name_handler()
        self.d2_array.d1_frame_list[-1].ix_min_handler()
        self.d2_array.d1_frame_list[-1].length_handler()
        self.d2_array.d1_frame_list[-1].data_source_handler()
        self.d2_array.d1_frame_list[-1].data_type_handler()

    def make_datum_frame(self, event=None):
        '''
        Adds an entry to self.data_dict if necessary, and updates
        self.datum_frame to show the values for the current datum
        '''
        ix = self.data_ix.get()
        if ix == 'PLEASE SPECIFY MIN/MAX INDICES':
            return
        ix = int(ix)
        if ix not in self.data_dict.keys():
            self.data_dict[ix] = {}
        self.datum_frame.destroy()
        self.datum_frame = tk.Frame(self)
        def datum_ttp(x):
            ''' Shortcut for commonly used construct '''
            return tk_tao_parameter(x, self.datum_frame, self.pipe)
        # Parameters
        param_list = ['data_source;ENUM;T;', 'data_type;DAT_TYPE;T;',
                'ele_name;STR;T;', 'ele_start_name;STR;T;', 'ele_ref_name;STR;T;',
                'data^merit_type;ENUM;T;', 'meas_value;REAL;T;', 'good_meas;LOGIC;T;T',
                'ref_value;REAL;T;', 'good_ref;LOGIC;T;T', 'weight;REAL;T;',
                'good_user;LOGIC;T;T', '1^ix_bunch;INUM;T;',
                'eval_point;ENUM;T;', 's_offset;REAL;T;']
        param_list = list(map(str_to_tao_param, param_list))
        # Fill in defaults set at d1 level
        param_list[0].value = self.d1_array_wids[1].tk_var.get()
        param_list[1].value = self.d1_array_wids[2].tk_var.get()
        param_list[5].value = self.d1_array_wids[3].tk_var.get()
        param_list[8].value = self.d1_array_wids[4].tk_var.get()
        param_list[9].value = bool(self.d1_array_wids[6].tk_var.get())
        # Default eval_point to End
        param_list[13].value = "End"
        param_list[13].value = self.d1_array_wids[5].tk_var.get()
        # Set parameter values if specified
        for i in range(len(param_list)):
            if param_list[i].name in self.data_dict[ix].keys():
                param_list[i].value = self.data_dict[ix][param_list[i].name]
        # Create widgets
        self.datum_wid_list = list(map(datum_ttp, param_list))
        def data_writer(i):
            ''' Writes the contents of self.datum_wid_list[i] into self.data_dict '''
            if self.datum_wid_list[i].param.type == 'LOGIC':
                val = bool(self.datum_wid_list[i].tk_var.get())
            else:
                val = self.datum_wid_list[i].tk_var.get()
            self.data_dict[ix][self.datum_wid_list[i].param.name] = val
        def data_writer_callback(i):
            ''' Callback for data_writer() '''
            return lambda *args : data_writer(i)

        # Set state of ele_names and s_offset for selected data_type
        def data_type_callback(*args):
            # ele_names
            if self.datum_wid_list[1]._has_ele():
                for i in range(2, 5):
                    self.datum_wid_list[i].tk_wid.configure(state='normal')
            else:
                for i in range(2, 5):
                    self.datum_wid_list[i].tk_wid.configure(state='disabled')
            # s_offset
            if self.datum_wid_list[1]._has_s_offset():
                self.datum_wid_list[11].tk_wid.configure(state='normal')
            else:
                self.datum_wid_list[11].tk_wid.configure(state='disabled')
        # Grid widgets
        for i in range(len(self.datum_wid_list)):
            self.datum_wid_list[i].tk_label.grid(row=i, column=0, sticky='W')
            self.datum_wid_list[i].tk_wid.grid(row=i, column=1, sticky='EW')
            # Write changes into self.data_dict
            self.datum_wid_list[i].tk_var.trace('w', data_writer_callback(i))
        self.datum_wid_list[1].tk_var.trace('w', data_type_callback)
        self.datum_frame.grid_columnconfigure(0, weight=1)
        self.datum_frame.grid_columnconfigure(1, weight=1)
        self.datum_frame.grid(row=2, column=4, rowspan=20, columnspan=2, sticky='NSEW')

    def datum_fill(self, ix):
        '''
        "Fills" the value in self.d1_array_wids[ix] to all datums in self.data_dict
        Actually just erases the values in self.data_dict, which produces the same effect
        '''
        for i in self.data_dict.keys():
            param = self.d1_array_wids[ix].param.name
            if param in self.data_dict[i].keys():
                self.data_dict[i].pop(param)
            elif 'data^'+param in self.data_dict[i].keys():
                self.data_dict[i].pop('data^'+param)
        # Redraw datum frame to reflect changes
        try:
            int(self.data_ix.get())
            self.make_datum_frame()
        except ValueError:
            pass

    def lat_browser(self, which):
        '''
        Opens a modified lattice table window to select elements en masse
        which should be either 'name', 'start_name', or 'ref_name'
        When the elements are chosen, ele_which will be set for each of this
        d1_array's data sequentially
        '''
        # Make sure ix_min and length are not invalidly set
        if (self.ix_min_handler() == False) or (self.length_handler() == False):
            return
        # Autosize the array if self.length == -1
        if self.length == -1:
            autosize = True
        else:
            autosize = False
        name = self.d2_array.name + '.' + self.name
        win = tao_ele_browser(self.d2_array.root, self.pipe, name, self,
                'data', which, self.d2_array.uni.get(), autosize=autosize)

    def name_handler(self, event=None, strict=False):
        '''
        Checks whether the input name is valid, and
        sets the tab name to it if it is
        Returns True if the name is valid,
        False if the name is already in use by another d1,
        and None if the name is empty
        If strict is set True, a warning will be displayed
        for an empty name
        '''
        if self.handler_block:
            return
        name = self.d1_array_wids[0].tk_var.get().strip()
        if name == "":
            self.name_warning_invalid.grid_forget()
            self.name_warning_empty.grid_forget()
            # Warning in strict mode
            if strict:
                self.name_warning_empty.grid(**self.name_warning_gs)
            self.name = self.BLANK_TITLE
            self.d2_array.notebook.tab(self.d2_array.d1_index, text=self.BLANK_TITLE)
            return None

        # Name is nonempty
        # -> check if it is invalid or already in use
        i = 0
        for d1 in self.d2_array.d1_frame_list:
            if (d1.name == name) & (self != self.d2_array.d1_frame_list[i]):
                self.name_warning_empty.grid_forget()
                self.name_warning_invalid.grid(**self.name_warning_gs)
                return False
            i = i+1

        # Name is nonempty and valid
        self.d2_array.notebook.tab(self.d2_array.d1_index, text=name)
        self.name_warning_empty.grid_forget()
        self.name_warning_invalid.grid_forget()
        if self.name != name:
            if self.d1_array_wids[2]._is_valid_dat_type(self.d2_array.name + '.' + name):
                # set data type
                self.d1_array_wids[2].tk_var.set(self.d2_array.name + '.' + name)
        self.name = name
        return True

    def ix_min_handler(self, event=None, strict=False):
        '''
        Checks if the starting index is valid,
        and updates self.ix_min if it is.
        Returns True if a good index is set,
        False if a bad index is set
        (non-integer or greater than ix_max),
        and None if the ix_min field is empty
        If strict is set true, prints a warning message
        for an empty ix_min
        '''
        if self.handler_block:
            return
        # Check if min index is empty
        ix_min = self.d1_array_wids[7].tk_var.get().strip()
        if ix_min == "":
            self.ix_min = -1
            self.ix_min_warning.grid_forget()
            if strict:
                self.ix_min_warning.grid(**self.ix_min_warning_gs)
            #print("ix_min empty")
            return None
        # Check if min index is a non-negative integer
        try:
            ix_min = int(ix_min)
        except ValueError:
            ix_min = -1
        if ix_min < 0:
            self.ix_min_warning.grid(**self.ix_min_warning_gs)
            self.ix_min = -1
            #print("ix_min invalid")
            return False
        # ix_min is a good value -> update self.ix_min
        #print("valid ix_min")
        self.ix_min = ix_min
        self.ix_min_warning.grid_forget()
        # Update the data index range if self.length is set
        if self.length != -1:
            #print("updating ix_max and data_chooser")
            self.ix_max = self.ix_min + self.length - 1
            self.data_chooser.configure(values=list(range(self.ix_min, self.ix_max+1)))
        return True

    def length_handler(self, event=None, strict=False):
        '''
        Checks that the d1 array length has been set
        to a valid value, and updates self.length if it has.
        Returns True if the length has been set to a valid
        (positive integer) value, False if set to an invalid
        value, and None if left blank.
        '''
        if self.handler_block:
            #print("handler blocked")
            return
        # Check if length is empty
        length = self.d1_array_wids[8].tk_var.get().strip()
        if length == "":
            self.length = -1
            self.length_warning.grid_forget()
            if strict:
                self.length_warning.grid(**self.length_warning_gs)
            #print("length empty")
            return None
        # Check if length is a positive integer
        try:
            length = int(length)
        except ValueError:
            length = -1
        if length < 1:
            self.length_warning.grid(**self.length_warning_gs)
            self.length = -1
            #print("length invalid")
            return False
        # length is a good value -> update self.length
        self.length = length
        self.length_warning.grid_forget()
        #print("length valid")
        #print("self.ix_min:" + str(self.ix_min))
        # Update the data index range if self.ix_min is set
        if self.ix_min != -1:
            #print("updating self.ix_max and data_chooser")
            self.ix_max = self.ix_min + self.length - 1
            self.data_chooser.configure(values=list(range(self.ix_min, self.ix_max+1)))
        return True

    def data_source_handler(self, *args):
        '''
        Updates the data_source widget to only show valid data types
        for the currently selected data source
        '''
        if self.handler_block:
            return
        source = self.d1_array_wids[1].tk_var.get()
        if source in ['beam', 'lat']:
            # Remake self.d1_array_wids[2] (data type)
            old_val = self.d1_array_wids[2].tk_var.get()
            self.d1_array_wids[2].tk_wid.grid_forget()
            self.d1_array_wids[2] = tk_tao_parameter(
                    str_to_tao_param('data_type;DAT_TYPE;T;'),
                    self, self.pipe, data_source=source)
            self.d1_array_wids[2].tk_var.set(old_val)
            self.d1_array_wids[2].tk_var.trace('w', self.data_type_handler)
            self.d1_array_wids[2].tk_wid.grid(row=6, column=1, sticky='EW')

    def data_type_handler(self, *args):
        '''
        Sets the state of the ele buttons appropriately for the currently selected data type
        '''
        if self.handler_block:
            return
        if not self.d1_array_wids[2]._has_ele():
            self.ele_name_button.configure(state='disabled')
            self.ele_start_name_button.configure(state='disabled')
            self.ele_ref_name_button.configure(state='disabled')
        else:
            self.ele_name_button.configure(state='normal')
            self.ele_start_name_button.configure(state='normal')
            self.ele_ref_name_button.configure(state='normal')




'''
This module defines the lat_element class, representing a single lattice
element in tao
'''
from .parameters import tao_parameter
from .parameters import tao_parameter_dict

class lat_element():
    '''
    Holds the essential information for a given lattice element.
    A lat_element has the following properties:
    self.id: the ele identifier in the form uni@branch>>ele_ix|which
    self.params: An ordered dictionary of tao_parameters
        This holds the general parameters related to the element
    self.has: An ordered dictionary of bools
        This holds whether or not this element has a given property
    self.num: An ordered dictionary of ints
        This holds the number of something this element has

    Init arguments:
    u_ix: the universe index of the desired element
    ix_branch: the branch index of the desired element
    ix_ele: the element index of the desired element
    which: one of 'base', 'model', or 'design'
    pipe: the tao_interface object to use for fetching the element info
            from tao (MUST be a tao_interface instance
            and not a tao_ctypes.Tao instance)
    '''
    def __init__(self, u_ix, ix_branch, ix_ele, which, pipe):
        self.pipe = pipe
        self.id = str(u_ix) + '@' + str(ix_branch) \
                + '>>' + str(ix_ele) + '|' + which
        fetch_cmd = "python ele:head " + self.id
        data_list = self.pipe.cmd_in(fetch_cmd)
        data_list = data_list.splitlines()

        p_list = [] # For parameters
        h_list = [] # For has#...
        n_list = [] # For num#...
        # All elements have gen_attribs:
        h_list.append("gen_attribs;LOGIC;F;T")
        for item in data_list:
            if item.find("has#") != -1:
                h_list.append(item[4:])
            elif item.find("num#") != -1:
                n_list.append(item[4:])
            else:
                # Manually set can_vary to F for base and design
                if which != "model":
                    item_parts = item.split(';')
                    item = ""
                    for i in range(len(item_parts)):
                        if i != 2:
                            item += item_parts[i] + ';'
                        else:
                            item += 'F;'
                    # Remove extra ;
                    item = item[:-1]
                p_list.append(item)

        self.params = tao_parameter_dict(p_list)
        self.has = tao_parameter_dict(h_list)
        self.num = tao_parameter_dict(n_list)



import math
local_funcs = {
    "acos": math.acos, 
    "acosh": math.acosh, 
    "asin": math.asin, 
    "asinh": math.asinh, 
    "atan": math.atan, 
    "atan2": math.atan2, 
    "atanh": math.atanh, 
    "ceil": math.ceil, 
    "cos": math.cos, 
    "cosh": math.cosh, 
    "degrees": math.degrees,
    "e_log": math.e,
    "erf": math.erf, 
    "erfc": math.erfc, 
    "exp": math.exp, 
    "factorial": math.factorial, 
    "floor": math.floor, 
    "gamma": math.gamma, 
    "log": math.log, 
    "log10": math.log10, 
    "pi": math.pi, 
    "pow": math.pow, 
    "radians": math.radians, 
    "sin": math.sin, 
    "sinh": math.sinh, 
    "sqrt": math.sqrt,
    "tan": math.tan,
    "tanh": math.tanh}

def eval_expr(expr):
  if "__" in expr or "=" in expr: raise  # Disallow unsafe evaluations.
  return eval(expr, local_funcs)



import numpy as np
import matplotlib as mp
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.path import Path
from matplotlib.backend_tools import ToolBase, ToolToggleBase
from matplotlib.widgets import Slider, Button, RadioButtons

from .tao_interface import *
from pytao.util.parameters import *

class taoplot:
    def __init__(self,pipe,PlotRegion):
        '''initializer, takes a tao interface and a graph region'''
        self.pipe = pipe #tao_interface object
        self.PlotRegion = PlotRegion #string describing region in tao of the desired plot

    ## @profile
    def plot(self, width):
        '''Returns a figure containing graphs using the data in the region PlotRegion of the tao instance in pipe, and plots a lat_layout below if applicable.
           Also returns information about the indices and locations of elements, width is used to modify the width of floor plan elements and the size of the lat layout.'''

        # Creates plotting figure
        fig = plt.figure()

        pipe = self.pipe # tao interface instance to plot in
        PlotRegion = self.PlotRegion # Region plot is in.

        # Determines if shapes are drawn on graph, changed to True later if needed
        LatLayout = False # draw a layout below data graphs?
        FloorPlan = False

        # Records information about element locations to be returned with the figure
        eleIndexList = []
        eleStartDict = {}
        eleEndDict = {}
        fpeIndexList = []
        fpeCenterDict = {}
        fpeRadiusDict = {}
        pathDict = {}

        def mpl_color(x):
            '''takes string containing pgplot color and returns corresponding matplotlib color'''
            x = x.lower()
            if x == 'yellow_green':
                return 'greenyellow'
            elif x == 'light_green':
                return 'limegreen'
            elif x == 'navy_blue':
                return 'navy'
            elif x == 'reddish_purple':
                return 'mediumvioletred'
            elif x == 'dark_grey':
                return 'gray'
            elif x == 'light_grey':
                return 'lightgray'
            elif x == 'Transparent' or x == 'transparent':
                return 'none'
            else:
                return x
        # End: def color

        def mpl_string(x):
            '''Takes string with pgplot characters and returns string with characters replaced with matplotlib equivalent.
               Raises NotImplementedError if an unknown pgplot character is used.'''
            x=x.replace('\\','\\\\')
            if '\\\\' in x:
                lx = '$'+x+'$'
                while lx.find('\\\\d') != -1 and lx.find('\\\\u') != -1:
                    if lx.find('\\\\d') <       lx.find('\\\\u'):
                        if lx.find('\\\\d') != -1:
                            start = lx.find('\\\\d')
                            end = lx.find('\\\\u')
                            sx = lx[start:end+3]
                            lx = lx.replace(sx,'_'+sx[3:-3])
                            sx = ''
                            start = ''
                    else:
                        if lx.find('\\\\u') != -1:
                            start = lx.find('\\\\u')
                            end = lx.find('\\\\d')
                            sx = lx[start:end+3]
                            lx = lx.replace(sx,'^'+sx[3:-3])
                            sx = ''
                            start = ''
                            end = ''

                lx = lx.replace(' ','\\ ')
                lx = lx.replace('%','\\%')

                lx = lx.replace('\\\\(2265)','\\partial')
                lx = lx.replace('\\\\ga','\\alpha')
                lx = lx.replace('\\\\gb','\\beta')
                lx = lx.replace('\\\\gg','\\gamma')
                lx = lx.replace('\\\\gd','\\delta')
                lx = lx.replace('\\\\ge','\\epsilon')
                lx = lx.replace('\\\\gz','\\zeta')
                lx = lx.replace('\\\\gy','\\eta')
                lx = lx.replace('\\\\gh','\\theta')
                lx = lx.replace('\\\\gi','\\iota')
                lx = lx.replace('\\\\gk','\\kappa')
                lx = lx.replace('\\\\gl','\\lambda')
                lx = lx.replace('\\\\gm','\\mu')
                lx = lx.replace('\\\\gn','\\nu')
                lx = lx.replace('\\\\gc','\\xi')
                lx = lx.replace('\\\\go','\\omicron')
                lx = lx.replace('\\\\gp','\\pi')
                lx = lx.replace('\\\\gr','\\rho')
                lx = lx.replace('\\\\gs','\\sigma')
                lx = lx.replace('\\\\gt','\\tau')
                lx = lx.replace('\\\\gu','\\upsilon')
                lx = lx.replace('\\\\gf','\\phi')
                lx = lx.replace('\\\\gx','\\chi')
                lx = lx.replace('\\\\gq','\\psi')
                lx = lx.replace('\\\\gw','\\omega')

                lx = lx.replace('\\\\gA','A')
                lx = lx.replace('\\\\gB','B')
                lx = lx.replace('\\\\gG','\\Gamma')
                lx = lx.replace('\\\\gD','\\Delta')
                lx = lx.replace('\\\\gE','E')
                lx = lx.replace('\\\\gZ','Z')
                lx = lx.replace('\\\\gY','H')
                lx = lx.replace('\\\\gH','\\Theta')
                lx = lx.replace('\\\\gI','I')
                lx = lx.replace('\\\\gK','\\Kappa')
                lx = lx.replace('\\\\gL','\\Lambda')
                lx = lx.replace('\\\\gM','M')
                lx = lx.replace('\\\\gN','N')
                lx = lx.replace('\\\\gC','\\Xi')
                lx = lx.replace('\\\\gO','O')
                lx = lx.replace('\\\\gP','\\Pi')
                lx = lx.replace('\\\\gR','P')
                lx = lx.replace('\\\\gS','\\Sigma')
                lx = lx.replace('\\\\gT','T')
                lx = lx.replace('\\\\gU','\\Upsilon')
                lx = lx.replace('\\\\gF','\\Phi')
                lx = lx.replace('\\\\gX','X')
                lx = lx.replace('\\\\gQ','\\Psi')
                lx = lx.replace('\\\\gW','\\Omega')
                lx = lx.replace('\\\\fn', '')
                if '\\\\' in lx:
                    raise NotImplementedError('unknown character in string, character not yet added to mpl_string: ' + repr(lx))
                return lx
            else:
                return x
        # End: def mpl_string(x)


        def circle_intersection(x1,y1,x2,y2,r):
            '''takes centers and radius of circles, returns the 2 intersection points of overlapping circles with equal radii'''
            dx = x2-x1
            dy = y2-y1
            d = np.sqrt(dx**2 + dy**2)
            a = d/2
            h = np.sqrt(r**2 - a**2)
            xm = x1 + dx/2
            ym = y1 + dy/2
            xs1 = xm + h*dy/d
            xs2 = xm - h*dy/d
            ys1 = ym - h*dx/d
            ys2 = ym + h*dx/d
            return (xs1,ys1),(xs2,ys2)


        def line(p1, p2):
            '''returns lines based on given points to be used with intersect'''
            A = (p1[1] - p2[1])
            B = (p2[0] - p1[0])
            C = (p1[0]*p2[1] - p2[0]*p1[1])
            return A, B, -C


        def intersect(L1, L2):
            '''returns intersection point of 2 lines from the line funciton, or false if the lines don't intersect'''
            D = L1[0]*L2[1] - L1[1]*L2[0]
            Dx = L1[2]*L2[1] - L1[1]*L2[2]
            Dy = L1[0]*L2[2] - L1[2]*L2[0]

            if D != 0:
                x = Dx / D
                y = Dy / D
                return x,y
            else:
                return False


        StylesDict = {
            'solid':'solid',
            'dashed':'dashed',
            'dash_dot':'dashdot',
            'dotted':'dotted',
            'dash_dot3':'dashdot', #currently the same as dashdot
            '1':'solid',
            '2':'dashed',
            '3':'dashdot',
            '4':'dotted',
            '5':'dashdot'
            }

        FillDict = {
            'solid_fill':'full',
            'no_fill':'none',
            'hatched':'full',
            'cross_hatched':'full',
            '1':'full',
            '2':'none',
            '3':'full',
            '4':'full'
            }

        #Dictionary with pgplot symbol strings as keys and corresponding matplotlib symbol strings as values
        SymbolsDict = {
            'do_not_draw':'',
            'square':'s', #no fill
            'dot':'.',
            'plus':'+',
            'times':(6,2,0),
            'circle':'$\\circ$',
            'x':'x',
            'x_symbol':'x',
            'triangle':'^', #no fill
            'circle_plus':'$\\oplus$',
            'circle_dot':'$\\odot$',
            'square_concave':(4,1,45),
            'diamond':'d', #no fill
            'star5':'*', #no fill
            'triangle_filled':'^', #fill
            'red_cross':'P',
            'star_of_david':(6,1,0),
            'square_filled':'s', #fill
            'circle_filled':'o', #fill
            'star5_filled':'*', #fill
            '0':'s', #no fill
            '1':'.',
            '2':'+',
            '3':(6,2,0),
            '4':'$\\circ$',
            '5':'x',
            '6':'s',
            '7':'^', #no fill
            '8':'$\\oplus$',
            '9':'$\\odot$',
            '10':(4,1,45),
            '11':'d', #no fill
            '12':'*', #no fill
            '13':'^', #fill
            '14':'P',
            '15':(6,1,0),
            '16':'s', #fill
            '17':'o', #fill
            '18':'*', #fill
            '-1':',',
            '-2':',',
            '-3':'(3,0,0)',
            '-4':'(4,0,0)',
            '-5':'(5,0,0)',
            '-6':'(6,0,0)',
            '-7':'(7,0,0)',
            '-8':'(8,0,0)',
            '-9':'(9,0,0)',
            '-10':'(10,0,0)',
            '-11':'(11,0,0)',
            '-12':'(12,0,0)'
        }

        # Plot Data...

        # List of plotting parameter strings from tao command python plot1
        pInfo = pipe.cmd_in('python plot1 '+PlotRegion,no_warn = True).splitlines()

        # Dictionary of Tao plot parameter name to the corresponding value.
        pInfoDict = {}
        for i in range(len(pInfo)):
            pInfoDict[pInfo[i].split(';')[0]] = str_to_tao_param(pInfo[i])

        # List of graph names and heights
        gNameList = []
        gHeightsList = []
        for i in range(pInfoDict['num_graphs'].value):
            gNameList.append(pInfoDict[('graph['+str(i+1)+']')].value)
            gHeightsList.append(1) #makes graphs the same size
        number_graphs = len(gNameList) + 1

        # Relative height of lat layout to graph if there is one graph
        layout_height = 0.2 * width

        # Makes layout approximately a fixed size as number of graphs changes
        gHeightsList.append(layout_height*len(gNameList) + layout_height*(len(gNameList)-1))

        gs = fig.add_gridspec(nrows = number_graphs,ncols = 1, height_ratios = gHeightsList)

        gSubPlotList = [None]  # List of graphs with the graph's matplotlib subplot as it's value. Indexed from 1.
        gFullNameList = [] # List of graph names, eg: r1.g or r3.x

        for gNum in range(1, len(gNameList)+1):

            # Create plots in figure, second line also makes x axes scale together
            if gNum == 1:
                gSubPlotList.append(fig.add_subplot(gs[gNum-1,0]))
            elif gNum > 1:
                gSubPlotList.append(fig.add_subplot(gs[gNum-1,0],sharex = gSubPlotList[1]))

            # Graph Data...

            # Graph "region.graph" full name, EG: "r13.g" or "top.x"
            gFullName = PlotRegion + '.' + gNameList[gNum-1]
            gFullNameList.append(gFullName)

            # List of graph parameters from tao command python plot_graph
            gInfo = pipe.cmd_in('python plot_graph '+gFullName,no_warn = True).splitlines()

            # Dictionary of graph parameters.
            gInfoDict = {}
            for i in range(len(gInfo)):
                gInfoDict[gInfo[i].split(';')[0]] = str_to_tao_param(gInfo[i])

            #------------------
            # Curve Data...

            # List of curve names
            cList = []
            for i in range(gInfoDict['num_curves'].value):
                cList.append(gInfoDict[('curve['+str(i+1)+']')].value)

            # List of curve parameters.
            cInfo = []
            for i in cList:
                cInfo.append(pipe.cmd_in('python plot_curve '+gFullName+'.'+i,no_warn = True).splitlines())

            cInfoDictList = []
            for i in range(len(cList)):
                cInfoDict = {}
                for j in range(len(cInfo[i])):
                    cInfoDict[cInfo[i][j].split(';')[0]] = str_to_tao_param(cInfo[i][j])
                cInfoDictList.append(cInfoDict)
                cInfoDict = {}


            #------------------
            # Line Data...

            # List of points from tao command python plot_line for each curve
            lInfo = []
            try:
                for i in cList:
                    lInfo.append(pipe.cmd_in('python plot_line '+gFullName+'.'+i,no_warn = True).splitlines())

                # List of lists of points used to draw each curve
                if len(lInfo) != 0:
                    PointsSuperList = []
                    for i in range(len(cList)):
                        pList = []
                        for j in range(len(lInfo[i])):
                            pList.append(lInfo[i][j].split(';'))
                        LineCoords = []
                        for j in range(len(pList)):
                            LineCoords.append([float(pList[j][1]),float(pList[j][2])])
                        PointsSuperList.append(LineCoords)
                        LineCoords = []
            except: #handle graph with no lines
                lInfo = []
                PointsSuperList = []
                for i in cList:
                    PointsSuperList.append([[0,0]])


            #------------------
            # Symbol Data

            # List  of points from tao command python plot_symbol for each curve
            try:
                sInfo = []
                for i in cList:
                    sInfo.append(pipe.cmd_in('python plot_symbol '+gFullName+'.'+i,no_warn = True).splitlines())
                SymbolSuperList = []
                # List of points used to draw symbols on each curve
                for i in range(len(cList)):
                    sList = []
                    for j in range(len(sInfo[i])):
                        sList.append(sInfo[i][j].split(';'))
                    SymCoords = []
                    for j in range(len(sList)):
                        SymCoords.append([float(sList[j][2]),float(sList[j][3])])
                    SymbolSuperList.append(SymCoords)
                    SymCoords = []
            except: # Handle graph with no symbols
                sInfo = []
                SymbolSuperList = []
                for i in cList:
                    SymbolSuperList.append([[0,0]])

            #------------------
            # Histogram Data...

            try:
                hInfo = []
                for i in cList:
                    hInfo.append(pipe.cmd_in('python plot_histogram '+gFullName+'.'+i,no_warn = True).splitlines())
                hInfoDictList = []
                # List of lists of dictionaries of plot_histogram data for each curve
                for i in range(len(cList)):
                    hInfoDict = {}
                    for j in range(len(hInfo[i])):
                        hInfoDict[hInfo[i][j].split(';')[0]] = str_to_tao_param(hInfo[i][j])
                    hInfoDictList.append(hInfoDict)
                    hInfoDict = {}

            except:
                hInfo = []

            # Plot Data

            # List of data needed to plot line and symbol graphs
            # Includes points, and line and symbol information for each curve
            CurvesList = []
            for i in range(len(cList)):
                CurveData = []
                CurveData.append(PointsSuperList[i]) #points for each curve
                CurveData.append(SymbolSuperList[i]) #symbols for each curve
                CurveData.append(mpl_color(cInfoDictList[i]['line'].get_component('color'))) #line color
                CurveData.append(StylesDict[cInfoDictList[i]['line'].get_component('pattern').lower()]) #line style
                if (cInfoDictList[i]['draw_line'].value == True): #line width if drawn
                    CurveData.append(cInfoDictList[i]['line'].get_component('width'))
                else:
                    CurveData.append(0)
                CurveData.append(mpl_color(cInfoDictList[i]['symbol'].get_component('color'))) #symbol color

                if cInfoDictList[i]['symbol'].get_component('type') == 'dot' or cInfoDictList[i]['symbol'].get_component('type') == '1': #determine if symbol should be filled
                    CurveData.append(cInfoDictList[i]['symbol'].get_component('color'))
                elif cInfoDictList[i]['symbol'].get_component('type')[-6:] == 'filled':
                    CurveData.append(cInfoDictList[i]['symbol'].get_component('color'))
                elif cInfoDictList[i]['symbol'].get_component('type')[:1] == '-':
                    CurveData.append(cInfoDictList[i]['symbol'].get_component('color'))
                elif (FillDict[cInfoDictList[i]['symbol'].get_component('fill_pattern')] == 'solid'):
                    CurveData.append(cInfoDictList[i]['symbol'].get_component('color'))
                else:
                    CurveData.append('none')

                if (cInfoDictList[i]['draw_symbols'].value == True) and SymbolsDict[cInfoDictList[i]['symbol'].get_component('type')] != '': #symbol size if drawn
                    CurveData.append(cInfoDictList[i]['symbol'].get_component('height'))
                else:
                    CurveData.append(0)

                if SymbolsDict[cInfoDictList[i]['symbol'].get_component('type')] != '':
                    CurveData.append(SymbolsDict[cInfoDictList[i]['symbol'].get_component('type')]) #symbol type
                else:
                    CurveData.append('.')
                CurveData.append(cInfoDictList[i]['symbol'].get_component('line_width')) #symbol line width
                CurvesList.append(CurveData)
                CurveData = []

            #------------------
            # Plotting...

            # Plots line graphs, histograms, phase space plots, dynamic aperture graphs, and plots for wave analysis
            LineList = []
            for i in CurvesList:
                PointsList = i[0]
                SymbolsList = i[1]
                xpList = []
                ypList = []
                for j in PointsList:
                    xpList.append(j[0])
                    ypList.append(j[1])
                xsList = []
                ysList = []
                for k in SymbolsList:
                    xsList.append(k[0])
                    ysList.append(k[1])
                try:
                    yMax = max(.5*max(max(ypList),max(ysList)),2*max(max(ypList),max(ysList)))
                    yMin = min(.5*min(min(ypList),min(ysList)),2*min(min(ypList),min(ysList)))
                except ValueError:
                    raise ValueError('no points found, make sure data is properly initialized')
                #boundaries for wave analysis rectangles

                if gInfoDict['graph^type'].value == 'data' or gInfoDict['graph^type'].value == 'wave.0' or \
                   gInfoDict['graph^type'].value == 'wave.a' or gInfoDict['graph^type'].value == 'wave.b':
                    LineList.append(gSubPlotList[gNum].plot(xpList,ypList,color = i[2],linestyle = i[3],linewidth = i[4]/2))
                    gSubPlotList[gNum].plot(xsList,ysList,color = i[5],linewidth = 0,markerfacecolor = i[6],markersize = i[7]/2,marker = i[8],mew = i[9]/2)

                    # Add lat_layout if x-axis is "s" (longitudinal postion).
                    if pInfoDict['x_axis_type'].value == 's':
                        LatLayout = True

                    # Wave region boundaries
                    if gInfoDict['graph^type'].value != 'data': #wave analysis rectangles
                        wInfo = pipe.cmd_in('python wave params',no_warn = True).splitlines()
                        a1 = float(wInfo[1].split(';')[3])
                        a2 = float(wInfo[2].split(';')[3])
                        b1 = float(wInfo[3].split(';')[3])
                        b2 = float(wInfo[4].split(';')[3])

                    if i[5] == 'blue' or i[5] == 'navy' or i[5] == 'cyan' or i[5] == 'green' or i[5] == 'purple':
                        waveColor = 'orange'
                    else:
                        waveColor = 'blue'
                    #wave analysis rectangle color

                    if gInfoDict['graph^type'].value == 'wave.0' or gInfoDict['graph^type'].value == 'wave.a':
                        gSubPlotList[gNum].add_patch(patches.Rectangle((a1,yMin),a2-a1,yMax-yMin,fill = False,color = waveColor))
                    if gInfoDict['graph^type'].value == 'wave.0' or gInfoDict['graph^type'].value == 'wave.b':
                        gSubPlotList[gNum].add_patch(patches.Rectangle((b1,yMin),b2-b1,yMax-yMin,fill = False,color = waveColor))
                #line and symbol graphs

                elif gInfoDict['graph^type'].value == 'dynamic_aperture':
                    LineList.append(gSubPlotList[gNum].plot(xpList,ypList,color = i[2],linestyle = i[3],linewidth = i[4]/2))
                    gSubPlotList[gNum].plot(xsList,ysList,color = i[5],linewidth = 0,markerfacecolor = i[6],markersize = i[7]/2,marker = i[8],mew = i[9]/2)
                #dynamic aperture graphs

                elif gInfoDict['graph^type'].value == 'phase_space':
                    if lInfo != []:
                        LineList.append(gSubPlotList[gNum].plot(xpList,ypList,color = i[2],linestyle = i[3],linewidth = i[4]/2))
                        gSubPlotList[gNum].plot(xsList,ysList,color = i[5],linewidth = 0,markerfacecolor = i[6],markersize = i[7]/2,marker = i[8],mew = i[9]/2)
                    else:
                        LineList.append(gSubPlotList[gNum].plot(xsList,ysList,color = i[5],linewidth = 0,markerfacecolor = i[6],markersize = i[7]/2,marker = i[8],mew = i[9]/2))
                #phase space graphs

                elif gInfoDict['graph^type'].value == 'histogram':
                    LineList.append(gSubPlotList[gNum].hist(xpList,bins = int(hInfoDictList[CurvesList.index(i)]['number'].value),weights = ypList,histtype = 'step',color = i[5]))
                #histogram

            if gInfoDict['graph^type'].value == 'key_table':
                raise NotImplementedError('key table is not available in the GUI')

            if gInfoDict['graph^type'].value == 'lat_layout':
                LatLayout = True
                gNameList = []
                plt.axis('off')
            #sets up lat layout plot

            if gInfoDict['graph^type'].value == 'floor_plan':
                FloorPlan = True
                gNameList = []
                plt.axis('off')
            #sets up floor plan plot

            if gInfoDict['draw_axes'].value == False:
                plt.axis('off')
            #hides axes if draw_axes is turned off

            if gInfoDict['why_invalid'].value != '':
                raise ValueError(gInfoDict['why_invalid'].value + ', make sure graph is properly initialized')

            #plots line and symbol graphs and histograms, lat layouts and floor plans are drawn later
            #LineList gives names of curves




            plt.title(mpl_string(gInfoDict['title'].value)+' '+gInfoDict['title_suffix'].value)
            #plot title

            LegendList = [] #legends for each graph
            LabelList = [] #labels in each legend
            try:
                for i in range(len(CurvesList)):
                    LegendList.append(LineList[i][0])
                    if mpl_string(cInfoDictList[i]['legend_text'].value) != '':
                        LabelList.append(mpl_string(cInfoDictList[i]['legend_text'].value))
                    elif mpl_string(cInfoDictList[i]['data_type'].value) == 'physical_aperture':
                        LabelList.append(mpl_string(cInfoDictList[i]['data_type'].value))
                    else:
                        LabelList.append('')
                #list of curves to be added to a legend and list of labels for each curve in the legend
            except IndexError:
                raise NotImplementedError('unknown graph type')


            if (gInfoDict['draw_curve_legend'].value == True and LabelList != ['']) and gInfoDict['graph^type'].value != 'lat_layout' and gInfoDict['graph^type'].value != 'floor_plan':
                gSubPlotList[gNum].legend(LegendList,LabelList)
            #plot legend

            plt.xlabel(mpl_string(gInfoDict['x'].get_component('label')))
            plt.ylabel(mpl_string(gInfoDict['y'].get_component('label')))
            #plot axis labels

            gSubPlotList[gNum].grid(gInfoDict['draw_grid'].value,which = 'major',axis = 'both')
            #plot grid

            plt.xlim(gInfoDict['x'].get_component('min'),gInfoDict['x'].get_component('max'))
            plt.ylim(gInfoDict['y'].get_component('min'),gInfoDict['y'].get_component('max'))
            #set axis limits

            gSubPlotList[gNum].set_axisbelow(True)
            #place graphs over grid lines

        #-----------------------
        # Plots lat layouts

        if LatLayout == True:
            if gInfoDict['graph^type'].value != 'lat_layout': #add space for lat layout below graph
                latLayoutSubPlot = fig.add_subplot(gs[len(gNameList),0],sharex = gSubPlotList[1])

            else: # Standalone lat layout graph
                latLayoutSubPlot = fig.add_subplot(len(gNameList)+1,1,len(gNameList)+1,sharex = gSubPlotList[1])

            # List of parameter strings from tao command python plot_graph
            layInfo = pipe.cmd_in('python plot_graph layout.g',no_warn = True).splitlines()

            # Tao_parameter object names from python plot_graph
            # Dictionary of tao_parameter name string keys to the corresponding tao_parameter object
            layInfoDict = {}
            for i in range(len(layInfo)):
                layInfoDict[layInfo[i].split(';')[0]] = str_to_tao_param(layInfo[i])



            # Makes lat layout only have horizontal axis for panning and zooming
            twinAxes = latLayoutSubPlot.axes.twinx()
            plt.xlim(gInfoDict['x'].get_component('min'),gInfoDict['x'].get_component('max'))
            plt.ylim(layInfoDict['y'].get_component('min'),layInfoDict['y'].get_component('max'))
            twinAxes.set_navigate(True)
            latLayoutSubPlot.axis('off')
            twinAxes.axis('off')

            # Sets axis limits and creates second axis to allow x panning and zooming
            latLayoutSubPlot.axes.set_navigate(False)
            latLayoutSubPlot.axhline(y = 0,xmin = 1.1*layInfoDict['x'].get_component('min'),xmax = 1.1*layInfoDict['x'].get_component('max'),color = 'Black')


            # Lat layout branch and universe information
            if layInfoDict['ix_universe'].value != -1:
                universe = layInfoDict[ix_universe].value
            else:
                universe = 1
            branch = layInfoDict['-1^ix_branch'].value


            # List of strings containing information about each element
            eleInfo = pipe.cmd_in('python plot_lat_layout '+str(universe)+'@'+str(branch),no_warn = True).splitlines()

            # All dict keys and entries are strings which match a lattice layout element index (eg: '1') string to the corresponding information
            eleIndexList = []
            eleStartDict = {} #element start coordinate
            eleEndDict = {} #element end coordinate
            eleLwDict = {} #element line width
            eleShapeDict = {} #element shape (eg: box)
            eleY1Dict = {} #height above 0
            eleY2Dict = {} #height below 0
            eleColorDict = {}
            eleNameDict = {}
            for i in range(len(eleInfo)):
                eleIndexList.append(int(eleInfo[i].split(';')[0]))
                eleStartDict[eleInfo[i].split(';')[0]] = float(eleInfo[i].split(';')[1])
                eleEndDict[eleInfo[i].split(';')[0]] = float(eleInfo[i].split(';')[2])
                eleLwDict[eleInfo[i].split(';')[0]] = float(eleInfo[i].split(';')[3])
                eleShapeDict[eleInfo[i].split(';')[0]] = eleInfo[i].split(';')[4].lower()
                eleY1Dict[eleInfo[i].split(';')[0]] = float(eleInfo[i].split(';')[5])
                eleY2Dict[eleInfo[i].split(';')[0]] = float(eleInfo[i].split(';')[6])
                eleColorDict[eleInfo[i].split(';')[0]] = mpl_color(eleInfo[i].split(';')[7].lower())
                eleNameDict[eleInfo[i].split(';')[0]] = eleInfo[i].split(';')[8]

            # Plotting line segments one-by-one can be slow if there are thousands of lattice elements.
            # So keep a list of line segments and plot all at once at the end.

            y_max = max(max(eleY1Dict.values()), max(eleY2Dict.values()))
            y2_floor = -max(eleY2Dict.values())  # Note negative sign
            lines = []
            widths = []
            colors = []

            for i in eleIndexList:
                s1 = eleStartDict[str(i)]
                s2 = eleEndDict[str(i)]
                y1 = eleY1Dict[str(i)]
                y2 = -eleY2Dict[str(i)]  # Note negative sign.
                wid = eleLwDict[str(i)]
                color = eleColorDict[str(i)]
                shape = eleShapeDict[str(i)]
                name = eleNameDict[str(i)]

                try:
                    # Normal case where element is not wrapped around ends of lattice.
                    if s2-s1 > 0:

                        # Draw box element
                        if shape == 'box':
                            latLayoutSubPlot.add_patch(patches.Rectangle((s1,y1), s2-s1, y2-y1, lw = wid,color = color,fill = False))

                        # Draw xbox element
                        elif shape == 'xbox':
                            latLayoutSubPlot.add_patch(patches.Rectangle((s1,y1), s2-s1, y2-y1, lw = wid,color = color,fill = False))
                            lines.extend([ [(s1,y1), (s2,y2)], [(s1,y2), (s2,y1)] ])
                            colors.extend([color, color])
                            widths.extend([wid, wid])

                        # Draw x element
                        elif shape == 'x':
                            lines.extend([ [(s1,y1), (s2,y2)], [(s1,y2), (s2,y1)] ])
                            colors.extend([color, color])
                            widths.extend([wid, wid])

                        # Draw bow_tie element
                        elif shape == 'bow_tie':
                            lines.extend([ [(s1,y1), (s2,y2)], [(s1,y2), (s2,y1)], [(s1,y1), (s1,y2)], [(s2,y1), (s2,y2)] ])
                            colors.extend([color, color, color, color])
                            widths.extend([wid, wid, wid, wid])

                        # Draw rbow_tie element
                        elif shape == 'rbow_tie':
                            lines.extend([ [(s1,y1), (s2,y2)], [(s1,y2), (s2,y1)], [(s1,y1), (s2,y1)], [(s1,y2), (s2,y2)] ])
                            colors.extend([color, color, color, color])
                            widths.extend([wid, wid, wid, wid])

                        # Draw diamond element
                        elif shape == 'diamond':
                            s_mid = (s1 + s2) / 2
                            lines.extend([ [(s1,0), (s_mid, y1)], [(s1,0), (s_mid, y2)], [(s2,0), (s_mid, y1)], [(s2,0), (s_mid, y2)] ])
                            colors.extend([color, color, color, color])
                            widths.extend([wid, wid, wid, wid])

                        # Draw circle element
                        elif shape == 'circle':
                            s_mid = (s1 + s2) / 2
                            latLayoutSubPlot.add_patch(patches.Ellipse((s_mid,0), y1-y2, y1-y2, lw = wid,color = color,fill = False))

                        # Draw element name
                        latLayoutSubPlot.text((s1+s2)/2, 1.1*y2_floor, name,ha = 'center',va = 'top',clip_on = True,color = color)

                    # Case where element is wrapped round the lattice ends.
                    else:
                        s_min = layInfoDict['x'].get_component('min')
                        s_max = layInfoDict['x'].get_component('max')

                        # Draw wrapped box element
                        if shape == 'box':
                            latLayoutSubPlot.plot([s1,s_max],[y1,y1],lw = wid,color = color)
                            latLayoutSubPlot.plot([s1,s_max],[y2,y2],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[y1,y1],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[y2,y2],lw = wid,color = color)
                            latLayoutSubPlot.plot([s1,s1],[y1,y2],lw = wid,color = color)
                            latLayoutSubPlot.plot([s2,s2],[y1,y2],lw = wid,color = color)

                        # Draw wrapped xbox element
                        elif shape == 'xbox':
                            latLayoutSubPlot.plot([s1,s_max],[y1,y1],lw = wid,color = color)
                            latLayoutSubPlot.plot([s1,s_max],[y2,y2],lw = wid,color = color)
                            latLayoutSubPlot.plot([s1,s_max],[y1,0],lw = wid,color = color)
                            latLayoutSubPlot.plot([s1,s_max],[y2,0],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[y1,y1],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[y2,y2],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[0,y1],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[0,y2],lw = wid,color = color)
                            latLayoutSubPlot.plot([s1,s1],[y1,y2],lw = wid,color = color)
                            latLayoutSubPlot.plot([s2,s2],[y1,y2],lw = wid,color = color)

                        # Draw wrapped x element
                        elif shape == 'x':
                            latLayoutSubPlot.plot([s1,s_max],[y1,0],lw = wid,color = color)
                            latLayoutSubPlot.plot([s1,s_max],[y2,0],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[0,y1],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[0,y2],lw = wid,color = color)

                        # Draw wrapped bow tie element
                        elif shape == 'bow_tie':
                            latLayoutSubPlot.plot([s1,s_max],[y1,y1],lw = wid,color = color)
                            latLayoutSubPlot.plot([s1,s_max],[y2,y2],lw = wid,color = color)
                            latLayoutSubPlot.plot([s1,s_max],[y1,0],lw = wid,color = color)
                            latLayoutSubPlot.plot([s1,s_max],[y2,0],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[y1,y1],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[y2,y2],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[0,y1],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[0,y2],lw = wid,color = color)

                        # Draw wrapped diamond element
                        elif shape == 'diamond':
                            latLayoutSubPlot.plot([s1,s_max],[0,y1],lw = wid,color = color)
                            latLayoutSubPlot.plot([s1,s_max],[0,y2],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[y1,0],lw = wid,color = color)
                            latLayoutSubPlot.plot([s_min,s2],[y2,0],lw = wid,color = color)

                        # Draw wrapped element name
                        latLayoutSubPlot.text(s_max, 1.1*y2_floor, name,ha = 'right',va = 'top',clip_on = True,color = color)
                        latLayoutSubPlot.text(s_min, 1.1*y2_floor, name,ha = 'left',va = 'top',clip_on = True,color = color)

                except KeyError:
                    pass

            # Draw all line segments
            latLayoutSubPlot.add_collection(mp.collections.LineCollection(lines, colors = colors, linewidths = widths))

            # Invisible line to give the lat layout enough vertical space.
            # Without this, the tops and bottoms of shapes could be cut off
            latLayoutSubPlot.plot([0, 0],[-1.7*y_max,1.3*y_max],lw = wid,color = color,alpha = 0)

        else:
            latLayoutSubPlot = fig.add_subplot(gs[len(gNameList),0],sharex = gSubPlotList[1])
            latLayoutSubPlot.remove()


        # Plots floor plans

        if FloorPlan == True:
            gSubPlotForFloorPlan = fig.add_subplot(len(gNameList)+1,1,len(gNameList)+1,sharex = gSubPlotList[1])

            floInfo = pipe.cmd_in('python plot_graph '+gFullName,no_warn = True).splitlines()
            #list of plotting parameter strings from tao command python plot_graph


            floInfoDict = {}
            for i in range(len(floInfo)):
                floInfoDict[floInfo[i].split(';')[0]] = str_to_tao_param(floInfo[i])
            #tao_parameter object names from python plot_graph for a floor plan
            #dictionary of tao_parameter name string keys to the corresponding tao_parameter object

            if floInfoDict['ix_universe'].value != -1:
                universe = floInfoDict[ix_universe].value

            else:
                universe = 1

            fpeInfo = pipe.cmd_in('python floor_plan '+gFullName,no_warn = True).splitlines()
            #list of plotting parameter strings from tao command python floor_plan


            fpeIndexList = [] # Contains lists of branch index then element index for each point
            fpeTypeDict = {}
            fpeSxDict = {} # Start x coordinate
            fpeSyDict = {} # Start y coordinate
            fpeSaDict = {} # Start angle
            fpeExDict = {} # End x coordinate
            fpeEyDict = {} # End y coordinate
            fpeEaDict = {} # End angle
            fpeLwDict = {} # Line width
            fpeShapeDict = {}
            fpeY1Dict = {} # Distance above
            fpeY2Dict = {} # Distance below
            fpeColorDict = {}
            fpeNameDict = {}
            fpeAlDict = {}  # Arc length
            fpeBaDict = {}  # Bend angle
            fpeSfaDict = {} # Relative angle of starting face to incoming line
            fpeEfaDict = {} # Relative angle of ending face to incoming line
            corner1 = {}    # Corner coordinates of objects
            corner2 = {}
            corner3 = {}
            corner4 = {}
            for i in range(len(fpeInfo)):
                fpeIndexList.append([int(fpeInfo[i].split(';')[0]),int(fpeInfo[i].split(';')[1])])
                fpeTypeDict[str(fpeIndexList[i])] = fpeInfo[i].split(';')[2].lower()
                fpeSxDict[str(fpeIndexList[i])] = float(fpeInfo[i].split(';')[3])
                fpeSyDict[str(fpeIndexList[i])] = float(fpeInfo[i].split(';')[4])
                fpeSaDict[str(fpeIndexList[i])] = float(fpeInfo[i].split(';')[5])
                fpeExDict[str(fpeIndexList[i])] = float(fpeInfo[i].split(';')[6])
                fpeEyDict[str(fpeIndexList[i])] = float(fpeInfo[i].split(';')[7])
                fpeEaDict[str(fpeIndexList[i])] = float(fpeInfo[i].split(';')[8])
                fpeLwDict[str(fpeIndexList[i])] = float(fpeInfo[i].split(';')[9])
                fpeShapeDict[str(fpeIndexList[i])] = fpeInfo[i].split(';')[10].lower()
                fpeY1Dict[str(fpeIndexList[i])] = width*float(fpeInfo[i].split(';')[11])
                fpeY2Dict[str(fpeIndexList[i])] = width*float(fpeInfo[i].split(';')[12])
                fpeColorDict[str(fpeIndexList[i])] = mpl_color(fpeInfo[i].split(';')[13].lower())
                fpeNameDict[str(fpeIndexList[i])] = fpeInfo[i].split(';')[14]
                try:
                    fpeAlDict[str(fpeIndexList[i])] = float(fpeInfo[i].split(';')[15])
                    fpeBaDict[str(fpeIndexList[i])] = float(fpeInfo[i].split(';')[16])
                    fpeSfaDict[str(fpeIndexList[i])] = float(fpeInfo[i].split(';')[17])
                    fpeEfaDict[str(fpeIndexList[i])] = float(fpeInfo[i].split(';')[18])
                except IndexError:
                    fpeAlDict[str(fpeIndexList[i])] = 0
                    fpeBaDict[str(fpeIndexList[i])] = 0
                    fpeSfaDict[str(fpeIndexList[i])] = 0
                    fpeEfaDict[str(fpeIndexList[i])] = 0

            #dict keys and entries are strings which match a floor plan element index (eg: '1') to the corresponding information

            conv = (180)/(np.pi) #radian to degree conversion
            for i in fpeIndexList:
                fpeCenterDict[str(i)] = ([fpeSxDict[str(i)] + (fpeExDict[str(i)]-fpeSxDict[str(i)])/2,fpeSyDict[str(i)] + (fpeEyDict[str(i)]-fpeSyDict[str(i)])/2])
                fpeRadiusDict[str(i)] = fpeY1Dict[str(i)] #for click detection
                ele_key = fpeTypeDict[str(i)]
                x1 = fpeSxDict[str(i)]
                x2 = fpeExDict[str(i)]
                y1 = fpeSyDict[str(i)]
                y2 = fpeEyDict[str(i)]
                off1 = fpeY1Dict[str(i)]
                off2 = fpeY2Dict[str(i)]
                width = fpeLwDict[str(i)]
                color = fpeColorDict[str(i)]
                angStart = fpeSaDict[str(i)]
                angEnd = fpeEaDict[str(i)]
                shape = fpeShapeDict[str(i)]
                ele_name = fpeNameDict[str(i)]
                relAngStart = fpeSfaDict[str(i)]
                relAngEnd = fpeEfaDict[str(i)]

                try:
                    if ele_key == 'drift' or ele_key == 'kicker':
                        gSubPlotForFloorPlan.plot([x1,x2],[y1,y2],color = 'black')
                    #draw drift element

                    if off1 == 0 and off2 == 0 and ele_key != 'sbend' and color != '':
                        gSubPlotForFloorPlan.plot([x1,x2],[y1,y2],lw = width,color = color)
                    #draw line element

                    elif shape == 'box' and ele_key != 'sbend' and color != '':
                        gSubPlotForFloorPlan.add_patch(patches.Rectangle((x1 + off2*np.sin(angStart), y1 - off2*np.cos(angStart)),
                                                       np.sqrt((x2-x1)**2 + (y2-y1)**2), off1+off2,
                                                       lw = width,color = color,fill = False,angle = angStart*conv))
                    #draw box element

                    elif shape == 'xbox' and ele_key != 'sbend' and color != '':
                        gSubPlotForFloorPlan.add_patch(patches.Rectangle((x1 + off2*np.sin(angStart), y1 - off2*np.cos(angStart)),
                                                                         np.sqrt((x2-x1)**2 + (y2-y1)**2), off1+off2,
                                                                         lw = width,color = color,fill = False,angle = angStart*conv))
                        gSubPlotForFloorPlan.plot([x1 + off2*np.sin(angStart),x2 - off1*np.sin(angStart)],[y1 - off2*np.cos(angStart),y2 + off1*np.cos(angStart)],lw = width,color = color)
                        gSubPlotForFloorPlan.plot([x1 - off1*np.sin(angStart),x2 + off2*np.sin(angStart)],[y1 + off1*np.cos(angStart),y2 - off2*np.cos(angStart)],lw = width,color = color)
                    #draw xbox element

                    elif shape == 'x' and ele_key != 'sbend' and color != '':
                        gSubPlotForFloorPlan.plot([x1 + off2*np.sin(angStart),x2 - off1*np.sin(angStart)],[y1 - off2*np.cos(angStart),y2 + off1*np.cos(angStart)],lw = width,color = color)
                        gSubPlotForFloorPlan.plot([x1 - off1*np.sin(angStart),x2 + off2*np.sin(angStart)],[y1 + off1*np.cos(angStart),y2 - off2*np.cos(angStart)],lw = width,color = color)
                    #draw x element


                    elif shape == 'bow_tie' and ele_key != 'sbend' and color != '':
                        gSubPlotForFloorPlan.plot([x1 + off2*np.sin(angStart),x2 - off1*np.sin(angStart)],[y1 - off2*np.cos(angStart),y2 + off1*np.cos(angStart)],lw = width,color = color)
                        gSubPlotForFloorPlan.plot([x1 - off1*np.sin(angStart),x2 + off2*np.sin(angStart)],[y1 + off1*np.cos(angStart),y2 - off2*np.cos(angStart)],lw = width,color = color)
                        gSubPlotForFloorPlan.plot([x1 - off1*np.sin(angStart),x2 - off1*np.sin(angStart)],[y1 + off1*np.cos(angStart),y2 + off1*np.cos(angStart)],lw = width,color = color)
                        gSubPlotForFloorPlan.plot([x1 + off2*np.sin(angStart),x2 + off2*np.sin(angStart)],[y1 - off2*np.cos(angStart),y2 - off2*np.cos(angStart)],lw = width,color = color)
                    #draw bow_tie element


                    elif shape == 'diamond' and ele_key != 'sbend' and color != '':
                        gSubPlotForFloorPlan.plot([x1,x1 + (x2-x1)/2 - off1*np.sin(angStart)],[y1,y1 + (y2-y1)/2 + off1*np.cos(angStart)],lw = width,color = color)
                        gSubPlotForFloorPlan.plot([x1 + (x2-x1)/2 - off1*np.sin(angStart),x2],[y1 + (y2-y1)/2 + off1*np.cos(angStart),y2],lw = width,color = color)
                        gSubPlotForFloorPlan.plot([x1,x1 + (x2-x1)/2 + off2*np.sin(angStart)],[y1,y1 + (y2-y1)/2 - off2*np.cos(angStart)],lw = width,color = color)
                        gSubPlotForFloorPlan.plot([x1 + (x2-x1)/2 + off2*np.sin(angStart),x2],[y1 + (y2-y1)/2 - off2*np.cos(angStart),y2],lw = width,color = color)
                    #draw diamond element


                    elif shape == 'circle' and ele_key != 'sbend' and color != '':
                        gSubPlotForFloorPlan.add_patch(patches.Circle((x1 + (x2-x1)/2,y1 + (y2-y1)/2),off1,lw = width,color = color,fill = False))
                    #draw circle element


                    elif shape == 'box' and ele_key == 'sbend' and color != '':
                        gSubPlotForFloorPlan.plot([x1-off1*np.sin(angStart-relAngStart), x1+off2*np.sin(angStart-relAngStart)],
                                                  [y1+off1*np.cos(angStart-relAngStart), y1-off2*np.cos(angStart-relAngStart)], lw = width, color = color)
                        gSubPlotForFloorPlan.plot([x2-off1*np.sin(angEnd+relAngEnd), x2+off2*np.sin(angEnd+relAngEnd)],
                                                  [y2+off1*np.cos(angEnd+relAngEnd), y2-off2*np.cos(angEnd+relAngEnd)], lw = width, color = color)
                        #draws straight sbend edges

                        intersection = intersect(line([x1-off1*np.sin(angStart), y1+off1*np.cos(angStart)], [x1+off2*np.sin(angStart), y1-off2*np.cos(angStart)]),
                                                 line([x2-off1*np.sin(angEnd), y2+off1*np.cos(angEnd)], [x2+off2*np.sin(angEnd), y2-off2*np.cos(angEnd+relAngEnd)]))
                        #center of circle used to draw arc edges of sbends

                        if intersection == False:
                            gSubPlotForFloorPlan.plot([x1-off1*np.sin(angStart-relAngStart), x2-off1*np.sin(angEnd+relAngEnd)],
                                                      [y1+off1*np.cos(angStart-relAngStart), y2+off1*np.cos(angEnd+relAngEnd)], lw = width, color = color)
                            gSubPlotForFloorPlan.plot([x1+off2*np.sin(angStart-relAngStart), x2+off2*np.sin(angEnd+relAngEnd)],
                                                      [y1-off2*np.cos(angStart-relAngStart), y2-off2*np.cos(angEnd+relAngEnd)], lw = width, color = color)
                        #draw sbend edges if bend angle is 0


                        elif intersection != False:
                            angle1 = 360 + conv*np.arctan2(y1+off1*np.cos(angStart-relAngStart)-intersection[1],x1-off1*np.sin(angStart-relAngStart)-intersection[0])
                            angle2 = 360 + conv*np.arctan2(y2+off1*np.cos(angEnd+relAngEnd)-intersection[1],x2-off1*np.sin(angEnd+relAngEnd)-intersection[0])
                            #angles of further curve endpoints relative to center of circle
                            angle3 = 360 + conv*np.arctan2(y1-off2*np.cos(angStart-relAngStart)-intersection[1],x1+off2*np.sin(angStart-relAngStart)-intersection[0])
                            angle4 = 360 + conv*np.arctan2(y2-off2*np.cos(angEnd+relAngEnd)-intersection[1],x2+off2*np.sin(angEnd+relAngEnd)-intersection[0])
                            #angles of closer curve endpoints relative to center of circle

                            if abs(angle1-angle2)<180:
                                a1 = min(angle1, angle2)
                                a2 = max(angle1, angle2)
                            else:
                                a1 = max(angle1, angle2)
                                a2 = min(angle1, angle2)

                            if abs(angle3-angle4)<180:
                                a3 = min(angle3, angle4)
                                a4 = max(angle3, angle4)
                            else:
                                a3 = max(angle3, angle4)
                                a4 = min(angle3, angle4)
                            #determines correct start and end angles for arcs

                            gSubPlotForFloorPlan.add_patch(patches.Arc((intersection[0],intersection[1]),
                                                            np.sqrt((x1-off1*np.sin(angStart-relAngStart)-intersection[0])**2 + (y1+off1*np.cos(angStart-relAngStart)-intersection[1])**2)*2,
                                                            np.sqrt((x1-off1*np.sin(angStart-relAngStart)-intersection[0])**2 + (y1+off1*np.cos(angStart-relAngStart)-intersection[1])**2)*2,
                                                            theta1 = a1,theta2 = a2,lw = width,color = color))
                            gSubPlotForFloorPlan.add_patch(patches.Arc((intersection[0],intersection[1]),
                                                            np.sqrt((x1+off2*np.sin(angStart-relAngStart)-intersection[0])**2 + (y1-off2*np.cos(angStart-relAngStart)-intersection[1])**2)*2,
                                                            np.sqrt((x1+off2*np.sin(angStart-relAngStart)-intersection[0])**2 + (y1-off2*np.cos(angStart-relAngStart)-intersection[1])**2)*2,
                                                            theta1 = a3,theta2 = a4,lw = width,color = color))
                            #draw sbend edges if bend angle is nonzero
                    #draw sbend element


                    if ele_name != '' and color != '' and np.sin(((angEnd+angStart)/2)) > 0:
                        gSubPlotForFloorPlan.text(x1+(x2-x1)/2 - 1.3*off1*np.sin(angStart), y1+(y2-y1)/2 + 1.3*off1*np.cos(angStart),
                                             ele_name, ha = 'right',va = 'center',color = 'black', rotation = -90+((angEnd+angStart)/2)*conv,clip_on = True,rotation_mode = 'anchor')

                    elif ele_name != '' and color != '' and np.sin(((angEnd+angStart)/2)) <= 0:
                        gSubPlotForFloorPlan.text(x1+(x2-x1)/2 - 1.3*off1*np.sin(angStart),y1+(y2-y1)/2 + 1.3*off1*np.cos(angStart),
                                             ele_name,ha = 'left',va = 'center',color = 'black',rotation = 90+((angEnd+angStart)/2)*conv,clip_on = True,rotation_mode = 'anchor')
                    #draw element name




                    '''floor plan click detection'''

                    if ele_key == 'sbend' and intersection != False: #for sbend click detection
                        c1 = [x1-off1*np.sin(angStart-relAngStart),y1+off1*np.cos(angStart-relAngStart)]
                        c2 = [x2-off1*np.sin(angEnd+relAngEnd),y2+off1*np.cos(angEnd+relAngEnd)]
                        c3 = [x1+off2*np.sin(angStart-relAngStart),y1-off2*np.cos(angStart-relAngStart)]
                        c4 = [x2+off2*np.sin(angEnd+relAngEnd),y2-off2*np.cos(angEnd+relAngEnd)]
                        #corners of sbend

                        if angStart > angEnd:
                            outerRadius = np.sqrt((x1-off1*np.sin(angStart-relAngStart)-intersection[0])**2 + (y1+off1*np.cos(angStart-relAngStart)-intersection[1])**2)
                            innerRadius = np.sqrt((x1+off2*np.sin(angStart-relAngStart)-intersection[0])**2 + (y1-off2*np.cos(angStart-relAngStart)-intersection[1])**2)
                        else:
                            outerRadius = -np.sqrt((x1-off1*np.sin(angStart-relAngStart)-intersection[0])**2 + (y1+off1*np.cos(angStart-relAngStart)-intersection[1])**2)
                            innerRadius = -np.sqrt((x1+off2*np.sin(angStart-relAngStart)-intersection[0])**2 + (y1-off2*np.cos(angStart-relAngStart)-intersection[1])**2)
                        #radii of sbend arc edges



                        middleAngle = (angStart+angEnd)/2

                        top = [intersection[0]-outerRadius*np.sin(middleAngle),intersection[1]+outerRadius*np.cos(middleAngle)]
                        bottom = [intersection[0]-innerRadius*np.sin(middleAngle),intersection[1]+innerRadius*np.cos(middleAngle)]
                        #midpoints of top and bottom arcs in an sbend

                        topCP = [2*(top[0])-.5*(c1[0])-.5*(c2[0]),2*(top[1])-.5*(c1[1])-.5*(c2[1])]
                        bottomCP = [2*(bottom[0])-.5*(c3[0])-.5*(c4[0]),2*(bottom[1])-.5*(c3[1])-.5*(c4[1])]
                        #corresponding control points for a quadratic Bezier curve that passes through the corners and arc midpoint

                        verts = [c1,topCP,c2,c4,bottomCP,c3,c1]
                        codes = [Path.MOVETO,Path.CURVE3,Path.CURVE3,Path.LINETO,Path.CURVE3,Path.CURVE3,Path.CLOSEPOLY]
                        pathDict[str(i)] = Path(verts,codes)

                        '''patch = patches.PathPatch(Path(verts,codes),facecolor = 'green',alpha = .5)
                        gSubPlotForFloorPlan.add_patch(patch)'''
                        #visualize clickable regions
                    #path approximating sbend region for clickable region on graph using lines and quadratic Bezier curves

                    else: #for non sbend click detection
                        corner1[str(i)] = [x1 - off1*np.sin(angStart),y1 + off1*np.cos(angStart)]
                        corner2[str(i)] = [x2 - off1*np.sin(angStart),y2 + off1*np.cos(angStart)]
                        corner3[str(i)] = [x1 + off2*np.sin(angStart),y1 - off2*np.cos(angStart)]
                        corner4[str(i)] = [x2 + off2*np.sin(angStart),y2 - off2*np.cos(angStart)]
                    #coordinates of corners of a floor plan element for clickable region


                except KeyError:
                    pass

                if gInfoDict['draw_axes'].value == False:
                    plt.axis('off')
                #hides axes if draw_axes is turned off


            '''Floor Plan Building Wall'''

            try:

                fbwInfo = pipe.cmd_in('python building_wall_graph '+gFullName,no_warn = True).splitlines()
                #list of plotting parameter strings from tao command python floor_building_wall

                fbwCurveList = []
                for i in range(len(fbwInfo)):
                    fbwCurveList.append(int(fbwInfo[i].split(';')[0]))

                fbwCurveList = list(set(fbwCurveList)) #list of unique curve indices

                bwn = pipe.cmd_in('python building_wall_list',no_warn = True).splitlines()
                bwnTypeDict = {}
                for i in range(len(bwn)):
                    bwnTypeDict[bwn[i].split(';')[0]] = bwn[i].split(';')[1]
                #dictionary where keys are wall indices and values are the corresponding building wall types

                fps = pipe.cmd_in('python shape_list floor_plan',no_warn = True).splitlines()
                fpsTypeDict = {} #building wall element types
                fpsColorDict = {} #building wall segment colors
                for i in range(len(fps)):
                    fpsTypeDict[fps[i].split(';')[1].split(':')[0].lower()] = fps[i].split(';')[2].lower()
                    if fps[i].split(';')[1].split(':')[0].lower() == 'building_wall':
                        fpsColorDict[fps[i].split(';')[1].split(':')[2].lower()] = mpl_color(fps[i].split(';')[3].lower())
                #dictionaries matching of the type building wall components to their color


                for i in fbwCurveList:
                    fbwIndexList = [] #index of point in curve
                    fbwXList = [] #list of point x coordinates
                    fbwYList = [] #list of point y coordinates
                    fbwRadiusList = [] #straight line if element has 0 or missing radius
                    for j in range(len(fbwInfo)):
                        if i == int(fbwInfo[j].split(';')[0]):
                            fbwIndexList.append(int(fbwInfo[j].split(';')[1]))
                            fbwXList.append(float(fbwInfo[j].split(';')[2]))
                            fbwYList.append(float(fbwInfo[j].split(';')[3]))
                            try:
                                fbwRadiusList.append(float(fbwInfo[j].split(';')[4]))
                            except:
                                fbwRadiusList.append(0.0)
                    k = max(fbwIndexList) #max line index

                    while k > 1:
                        kIndex = fbwIndexList.index(k)
                        mIndex = fbwIndexList.index(k-1) #adjacent point to connect to
                        if bwnTypeDict[str(i)] not in fpsColorDict.keys():
                            # TODO: This is a temporary fix to deal with building wall segments
                            # that don't have an associated floor_plan shape
                            # Currently this will fail to match to wild cards
                            # in the shape name (e.g. building_wall::* should match
                            # to every building wall segment, but currently it
                            # matches to none).  A more sophisticated way of getting the
                            # floor_plan shape settings for building walls will be required
                            # in the future, either through a python command in tao or
                            # with a method on the python to match wild cards to wall segment names
                            print("No floor_plan shape defined for building_wall segment " + bwnTypeDict[str(i)])
                            k -= 1
                            continue

                        if fbwRadiusList[kIndex] == 0: #draw building wall line
                            gSubPlotForFloorPlan.plot([fbwXList[kIndex],fbwXList[mIndex]],[fbwYList[kIndex],fbwYList[mIndex]],color = fpsColorDict[bwnTypeDict[str(i)]])

                        else: #draw building wall arc
                            centerList = circle_intersection(fbwXList[mIndex],fbwYList[mIndex],fbwXList[kIndex],fbwYList[kIndex],abs(fbwRadiusList[kIndex]))
                            #radius and endpoints specify 2 possible circle centers for arcs
                            mpx = (fbwXList[mIndex] + fbwXList[kIndex])/2
                            mpy = (fbwYList[mIndex] + fbwYList[kIndex])/2
                            if np.arctan2((fbwYList[mIndex]-mpy),(fbwXList[mIndex]-mpx)) < np.arctan2(centerList[0][1],centerList[0][0]) < np.arctan2((fbwYList[mIndex]-mpy),(fbwXList[mIndex]-mpx)) and fbwRadiusList[kIndex]>0:
                                center = (centerList[1][0],centerList[1][1])
                            elif np.arctan2((fbwYList[mIndex]-mpy),(fbwXList[mIndex]-mpx)) < np.arctan2(centerList[0][1],centerList[0][0]) < np.arctan2((fbwYList[mIndex]-mpy),(fbwXList[mIndex]-mpx)) and fbwRadiusList[kIndex]<0:
                                center = (centerList[0][0],centerList[0][1])
                            elif fbwRadiusList[kIndex]>0:
                                center = (centerList[0][0],centerList[0][1])
                            else:
                                center = (centerList[1][0],centerList[1][1])
                            #find correct center

                            mAngle = 360 + conv*np.arctan2((fbwYList[mIndex]-center[1]),(fbwXList[mIndex]-center[0]))
                            kAngle = 360 + conv*np.arctan2((fbwYList[kIndex]-center[1]),(fbwXList[kIndex]-center[0]))

                            if abs(kAngle-mAngle) <= 180:
                                if kAngle > mAngle:
                                    t1 = mAngle
                                    t2 = kAngle
                                else:
                                    t1 = kAngle
                                    t2 = mAngle
                            else:
                                if kAngle > mAngle:
                                    t1 = kAngle
                                    t2 = mAngle
                                else:
                                    t1 = mAngle
                                    t2 = kAngle
                            #pick correct start and end angle for arc

                            gSubPlotForFloorPlan.add_patch(patches.Arc(center,fbwRadiusList[kIndex]*2,fbwRadiusList[kIndex]*2,theta1 = t1,theta2 = t2,color = fpsColorDict[bwnTypeDict[str(i)]]))
                            #draw building wall arc

                        k = k - 1
            except ValueError:
                pass
            #plot floor plan building walls




            '''Floor Plan Orbit'''

            if float(floInfoDict['floor_plan_orbit_scale'].value) != 0:
                fpoInfo = pipe.cmd_in('python floor_orbit '+gFullName,no_warn = True).splitlines()

                fpoIndexList = []
                fpoXList = []
                fpoYList = []
                for i in range(len(fpoInfo)):
                    fpoIndexList.append(int(fpoInfo[i].split(';')[1]))

                    if fpoInfo[i].split(';')[2].lower() == 'x':
                        for j in range(3,len(fpoInfo[i].split(';'))):
                            fpoXList.append(float(fpoInfo[i].split(';')[j]))
                    if fpoInfo[i].split(';')[2].lower() == 'y':
                        for j in range(3,len(fpoInfo[i].split(';'))):
                            fpoYList.append(float(fpoInfo[i].split(';')[j]))

                gSubPlotForFloorPlan.plot(fpoXList,fpoYList,color = floInfoDict['floor_plan_orbit_color'].value.lower())
            #Lists of floor plan orbit point indices, x coordinates, and y coordinates
            #plot floor plan orbit




            '''floor plan labels and axes'''

            plt.xlabel(mpl_string(gInfoDict['x'].get_component('label')))
            plt.ylabel(mpl_string(gInfoDict['y'].get_component('label')))
            #plot floor plan axis labels

            gSubPlotForFloorPlan.grid(gInfoDict['draw_grid'].value,which = 'major',axis = 'both')
            plt.xlim(gInfoDict['x'].get_component('min'),gInfoDict['x'].get_component('max'))
            plt.ylim(gInfoDict['y'].get_component('min'),gInfoDict['y'].get_component('max'))
            gSubPlotForFloorPlan.set_axisbelow(True)
            #plot floor plan grid




        '''''''''Output Data'''''''''
        #creates list of data to be returned with the figure, needed to make elements clickable on graphs

        if gInfoDict['graph^type'].value == 'lat_layout' or gInfoDict['graph^type'].value == 'floor_plan':
            if gInfoDict['ix_universe'].value != -1:
                gUniverse = gInfoDict[ix_universe].value

            else:
                gUniverse = 1

            gBranch = gInfoDict['-1^ix_branch'].value
            gComponent = 'model'

        else:
            if gInfoDict['ix_universe'].value != -1:
                gUniverse = gInfoDict[ix_universe].value

            else:
                gUniverse = 1

            gBranch = gInfoDict['-1^ix_branch'].value
            if gInfoDict['component'].value == 'model' or gInfoDict['component'].value == 'base' or gInfoDict['component'].value == 'design':
                gComponent = gInfoDict['component'].value
            else:
                gComponent = 'model'
        #get universe, branch, and component

        if gInfoDict['graph^type'].value != 'floor_plan':
            corner1 = []
            corner2 = []
            corner3 = []
            corner4 = []
            fpeIndexList = []
            fpeShapeDict = []
            fpeCenterDict = []
            fpeRadiusDict = []
        if LatLayout != True:
            eleIndexList = []
            eleStartDict = []
            eleEndDict = []
            eleShapeDict = []
            eleY1Dict = []
        #fills output list with blank lists if information does not apply to the selected graph type

        returnList = [gInfoDict['graph^type'].value, gUniverse, gBranch, gComponent, eleIndexList, eleStartDict, eleEndDict, fpeIndexList,fpeShapeDict,fpeCenterDict, fpeRadiusDict, corner1, corner2, corner3, corner4, pathDict, eleShapeDict, eleY1Dict, gFullNameList]
        #data to be returned with the figure to make elements clickable

        fig.tight_layout(pad = .5) #prevents graphs from overlapping

        return fig, returnList



'''

'''





def create_d2_data_cmd(d2_name, d1_names, upper_bounds=[], lower_bounds=None):
    """
    Helper function to create a d2_data with several d1_data.
    
    Returns the command string for use in Tao. 
    
    Example usage:
        cmd = create_d2_data_cmd('X', ['A', 'B'], [2,4] )
        tao.cmd(cmd)
    """

    n = len(d1_names)
    if not lower_bounds:
        lower_bounds = n*[1]
    
    cmd = f'python data_d2_create {d2_name}^^{n}'
    for name, lower_bound, upper_bound in zip(d1_names, lower_bounds, upper_bounds):
        cmd+= f'^^{name}^^{lower_bound}^^{upper_bound}'
        
    return cmd



def create_datum_cmd(
 
    datum_name='d2.d1[1]',
    data_type='beta.a',
    ele_ref_name='',
    ele_start_name='',
    ele_name='end',
    merit_type='target',
    meas=0,
    good_meas=False,
    ref=0,
    good_ref=False,
    weight=0,
    good_user=True,
    data_source='',
    eval_point='end',
    s_offset=0,
    ix_bunch=0,
    invalid_value=0,
    spin_n0_x=0,
    spin_n0_y=0,
    spin_n0_z=0 
                ):
    """
    Helper function to create a datum, with defaults.
    
    Returns the command string for use in Tao. 
    
    Example usage:
        cmd=create_datum_cmd('X.A[1]', meas=1, good_meas=True) 
        tao.cmd(cmd)
        tao.cmd('python data_set_design_value') # Needed to initialize
        
    """
    # Split for readablity 
    cmd = 'python datum_create '\
            f'{datum_name}^^{data_type}^^{ele_ref_name}^^{ele_start_name}^^{ele_name}^^{merit_type}'\
            f'^^{meas}^^{good_meas}^^{ref}^^{good_ref}^^{weight}^^{good_user}^^{data_source}^^{eval_point}'\
            f'^^{s_offset}^^{ix_bunch}^^{invalid_value}^^{spin_n0_x}^^{spin_n0_y}^^{spin_n0_z}'
    
    return cmd




def create_variable_cmd(
    var_name='',
    ele_name='',
    attribute='',
    universes='1',
    weight=0,
    step=0,
    low_lim=0,
    high_lim=0,
    merit_type='target',
    good_user=True,
    key_bound=False,
    key_delta=1
):
    """
    TODO: this doesn't work. 
    
    Create a single variable
    Command syntax:
    python var_create {var_name}^^{ele_name}^^{attribute}^^{universes}^^{weight}^^{step}^^{low_lim}^^{high_lim}^^
                                                                         {merit_type}^^{good_user}^^{key_bound}^^{key_delta}
    
    """
    
    
    # Split for readablity 
    cmd = 'python var_create '\
        f'{var_name}^^{ele_name}^^{attribute}^^{universes}'\
        f'^^{weight}^^{step}^^{low_lim}^^{high_lim}'\
        f'^^{merit_type}^^{good_user}^^{key_bound}^^{key_delta}'
    return cmd